#ifndef OSP_CALIB_EXTRINSIC_H
#define OSP_CALIB_EXTRINSIC_H

#include <boost/property_tree/ptree.hpp>
#include <osp/core/utility.h>
#include <osp/exports/calib_export.h>
#include <osp/types/coordinates.h>
#include <osp/types/mat.h>
#include <osp/types/ndarray.h>

namespace osp {

class OSP_CALIB_EXPORT Extrinsic
{
public:
    Extrinsic();
    virtual ~Extrinsic() = default;

    static Extrinsic readFromFile(const std::string &filename);
    void saveToFile(const std::string &filename) const;

    Mat<double, 4, 4> H() const;
    Mat<double, 3, 3> rotMat() const;
    Vec3<double> rodrigues() const;
    Vec3<double> t() const;
    Vec3<double> c() const;

    void setH(Mat<double, 4, 4> H);
    void setRotMat(Mat<double, 3, 3> rotMat);
    void setRodrigues(Vec3<double> rodrigues);
    void setT(Vec3<double> t);
    void setC(Vec3<double> t);

private:
    Mat<double, 3, 3> m_rotMat;
    Vec3<double> m_t;
};

OSP_CALIB_EXPORT bool operator==(const Extrinsic &lhs, const Extrinsic &rhs);
OSP_CALIB_EXPORT bool operator!=(const Extrinsic &lhs, const Extrinsic &rhs);

template <typename T>
Point3d<T> camToWrld(const Point3d<T> &camPt, const Extrinsic &extrinsic)
{
    auto t = extrinsic.t();
    auto xt = camPt.x - static_cast<T>(t[0]);
    auto yt = camPt.y - static_cast<T>(t[1]);
    auto zt = camPt.z - static_cast<T>(t[2]);

    auto R = extrinsic.rotMat();
    return {
        static_cast<T>(R[0])*xt + static_cast<T>(R[3])*yt + static_cast<T>(R[6])*zt,
        static_cast<T>(R[1])*xt + static_cast<T>(R[4])*yt + static_cast<T>(R[7])*zt,
        static_cast<T>(R[2])*xt + static_cast<T>(R[5])*yt + static_cast<T>(R[8])*zt
    };
}

template <typename T>
Spherical<T> camToWrld(const Spherical<T> &camPt, const Extrinsic &extrinsic)
{
    return camToWrld(Point3d<T>{camPt}, extrinsic);
}

template <typename T>
Point3d<T> wrldToCam(const Point3d<T> &wrldPt, const Extrinsic &extrinsic)
{
    auto t = extrinsic.t();
    auto R = extrinsic.rotMat();
    return {
		static_cast<T>(R[0]) * wrldPt.x + static_cast<T>(R[1]) * wrldPt.y + static_cast<T>(R[2]) * wrldPt.z + static_cast<T>(t[0]),
		static_cast<T>(R[3]) * wrldPt.x + static_cast<T>(R[4]) * wrldPt.y + static_cast<T>(R[5]) * wrldPt.z + static_cast<T>(t[1]),
		static_cast<T>(R[6]) * wrldPt.x + static_cast<T>(R[7]) * wrldPt.y + static_cast<T>(R[8]) * wrldPt.z + static_cast<T>(t[2])
    };
}

template <typename T>
Spherical<T> wrldToCam(const Spherical<T> &wrldPt, const Extrinsic &extrinsic)
{
    return wrldToCam(Point3d<T>{wrldPt}, extrinsic);
}

template <typename T>
Matrix<T> camToWrld(const Matrix<T> &camPts, const Extrinsic &extrinsic)
{
    Matrix<T> wrldPts;
    processElementwise(camPts, wrldPts, [&](const T &src) { return camToWrld(src, extrinsic); });
    return wrldPts;
}

template <typename T>
Matrix<T> wrldToCam(const Matrix<T> &wrldPts, const Extrinsic &extrinsic)
{
    Matrix<T> camPts;
    processElementwise(wrldPts, camPts, [&](const T &src) { return wrldToCam(src, extrinsic); });
    return camPts;
}


OSP_CALIB_EXPORT bool readFromNode(Extrinsic &extrinsic, const boost::property_tree::ptree &node);
OSP_CALIB_EXPORT void writeToNode(const Extrinsic &extrinsic, boost::property_tree::ptree &node);

}


#endif /* OSP_CALIB_EXTRINSIC_H */
