#ifndef OSP_CALIB_INTRINSIC_H
#define OSP_CALIB_INTRINSIC_H

#include <boost/property_tree/ptree.hpp>
#include <boost/variant.hpp>
#include <osp/core/utility.h>
#include <osp/calib/projectionmodel.h>
#include <osp/exports/calib_export.h>
#include <osp/types/mat.h>
#include <osp/types/ndarray.h>

namespace osp {

class OSP_CALIB_EXPORT Intrinsic
{
public:
    Intrinsic();
    virtual ~Intrinsic() = default;

    static Intrinsic readFromFile(const std::string &filename);
    void saveToFile(const std::string &filename) const;

    Intrinsic(const Intrinsic &other);
    Intrinsic(Intrinsic && other);
    Intrinsic & operator = (Intrinsic other);

    friend void swap(Intrinsic &a, Intrinsic &b);

    const ProjectionModel &projectionModel() const;

    ProjectionModel &projectionModel();
    size_t imageWidth() const;
    size_t imageHeight() const;
    double fx() const;
    double fy() const;
    double cx() const;
    double cy() const;
    double skew() const;

    void setProjectionModel(ProjectionModel projectionModel);
    void setImageWidth(size_t imgWidth);
    void setImageHeight(size_t imgHeight);
    void setFx(double fx);
    void setFy(double fy);
    void setCx(double cx);
    void setCy(double cy);
    void setSkew(double skew);

    void setHFov(double hFov);
    void setVFov(double vFov);

    template <typename T = float>
    Matrix<Point2d<T>> imagePoints() const;

private:
    ProjectionModel m_projectionModel;
    size_t m_imgWidth;
    size_t m_imgHeight;
    double m_fx;
    double m_fy;
    double m_cx;
    double m_cy;
    double m_skew;
};

OSP_CALIB_EXPORT bool operator==(const Intrinsic &lhs, const Intrinsic &rhs);
OSP_CALIB_EXPORT bool operator!=(const Intrinsic &lhs, const Intrinsic &rhs);

OSP_CALIB_EXPORT bool readFromNode(Intrinsic &intrinsic, const boost::property_tree::ptree &node);
OSP_CALIB_EXPORT void writeToNode(const Intrinsic &intrinsic, boost::property_tree::ptree &node);

// implementation

template <typename T>
Matrix<Point2d<T>> Intrinsic::imagePoints() const
{
    Matrix<Point2d<T>> imgPts(m_imgHeight, m_imgWidth);

    for (size_t y = 0; y < m_imgHeight; ++y)
    {
        Point2d<T> *ptr = imgPts.ptr(y);

        for (size_t x = 0; x < m_imgWidth; ++x)
        {
            ptr->x = x;
            ptr->y = y;
            ptr += imgPts.stride(1);
        }
    }
    return imgPts;
}

template <typename T>
Point2d<T> imgNormToImg(const Point2d<T> &normImgPt, const Intrinsic &intrinsic)
{
    return {
        normImgPt.x * static_cast<T>(intrinsic.fx()) + normImgPt.y * static_cast<T>(intrinsic.skew()) + static_cast<T>(intrinsic.cx()),
        normImgPt.y * static_cast<T>(intrinsic.fy()) + static_cast<T>(intrinsic.cy())
    };
}

template <typename T>
Polar<T> imgNormToImg(const Polar<T> &normImgPt, const Intrinsic &intrinsic)
{
    return imgNormToImg(intrinsic, Point2d<T>{normImgPt});
}

template <typename SrcType>
Matrix<SrcType> imgNormToImg(const Matrix<SrcType> &imgNormPts, const Intrinsic &intrinsic)
{
    Matrix<SrcType> imgPts;
    processElementwise(imgNormPts, imgPts, [&](const SrcType &src) { return imgNormToImg(src, intrinsic); });
    return imgPts;
}

template <typename T>
Point2d<T> camToImg(const Point3d<T> &camPt, const Intrinsic &intrinsic)
{
    return imgNormToImg(project(intrinsic.projectionModel(), camPt), intrinsic);
}

template <typename T>
Polar<T> camToImg(const Spherical<T> &camPt, const Intrinsic &intrinsic)
{
    return imgNormToImg(project(intrinsic.projectionModel(), camPt));
}

template <typename T>
Point2d<T> imgToImgNorm(const Point2d<T> &imgPt, const Intrinsic &intrinsic)
{
    Point2d<T> imgNormPt;
    imgNormPt.y = (imgPt.y - static_cast<T>(intrinsic.cy())) / static_cast<T>(intrinsic.fy());
    imgNormPt.x = (imgPt.x - static_cast<T>(intrinsic.skew())*imgNormPt.y - static_cast<T>(intrinsic.cx())) / static_cast<T>(intrinsic.fx());
    return imgNormPt;
}

template <typename T>
Polar<T> imgToImgNorm(const Polar<T> &imgPt, const Intrinsic &intrinsic)
{
    return imgToImgNorm(intrinsic, Point2d<T>{imgPt});
}

template <typename SrcType>
Matrix<SrcType> imgToImgNorm(const Matrix<SrcType> &imgPts, const Intrinsic &intrinsic)
{
    Matrix<SrcType> imgNormPts;
    processElementwise(imgPts, imgNormPts, [&](const SrcType &src) { return imgToImgNorm(src, intrinsic); });
    return imgNormPts;
}

template <typename T>
typename Reprojected<T>::type imgToCam(const T &imgPt, const Intrinsic &intrinsic, typename T::ValueType norm = 1)
{
    return reproject(intrinsic.projectionModel(), imgToImgNorm(imgPt, intrinsic)) * norm;
}

template <typename SrcType>
Matrix<typename Reprojected<SrcType>::type> imgToCam(const Matrix<SrcType> &imgPts, const Intrinsic &intrinsic, typename SrcType::ValueType norm = 1)
{
    using DstType = typename Reprojected<SrcType>::type;

    Matrix<DstType> camPts;
    processElementwise(imgPts, camPts, [&](const SrcType &src) { return imgToCam(src, intrinsic, norm); });
    return camPts;
}

template <typename SrcType>
Matrix<typename Projected<SrcType>::type> camToImg(const Matrix<SrcType> &camPts, const Intrinsic &intrinsic)
{
    using DstType = typename Projected<SrcType>::type;

    Matrix<DstType> imgPts;
    processElementwise(camPts, imgPts, [&](const SrcType &src) { return camToImg(src, intrinsic); });
    return imgPts;
}

}


#endif /* OSP_CALIB_INTRINSIC_H */
