#ifndef OSP_CALIB_PROJECTIONMODEL_H
#define OSP_CALIB_PROJECTIONMODEL_H

#include <boost/variant.hpp>
#include <osp/calib/projection/epipolarequidistant.h>
#include <osp/calib/projection/equiangular.h>
#include <osp/calib/projection/genericasymmetric.h>
#include <osp/calib/projection/genericradsym.h>
#include <osp/calib/projection/pinhole.h>
#include <osp/types/coordinates.h>
#include <osp/types/mat.h>
#include <osp/types/ndarray.h>

namespace osp {

using ProjectionVariant = boost::variant<
    EpipolarEquidistantProjection,
    EquiangularProjection,
    GenericAsymmetricProjection,
    GenericRadialSymmetricProjection,
    PinholeProjection
>;

class OSP_CALIB_EXPORT ProjectionModel
{
public:
    ProjectionModel();
    template <typename ProjectionType>
    ProjectionModel(ProjectionType projection);
    ~ProjectionModel() = default;

    ProjectionModel(const ProjectionModel &other);
    ProjectionModel &operator=(ProjectionModel other);
    friend void swap(ProjectionModel &a, ProjectionModel &b);

    ProjectionModelType type() const;
    std::string typeStr() const;

    template <typename T>
    void setProjection(T projection);

    template <typename T>
    T &projection();

    template <typename T>
    const T &projection() const;

    const ProjectionVariant &variant() const;
    ProjectionVariant &variant();

private:
    ProjectionVariant m_projection;
};

OSP_CALIB_EXPORT bool operator==(const ProjectionModel &lhs, const ProjectionModel &rhs);
OSP_CALIB_EXPORT bool operator!=(const ProjectionModel &lhs, const ProjectionModel &rhs);

template <typename ProjectionType>
ProjectionModel::ProjectionModel(ProjectionType proj)
    : m_projection(proj)
{
}

template <typename T>
void ProjectionModel::setProjection(T proj)
{
    m_projection = proj;
}

template <typename T>
T &ProjectionModel::projection()
{
    return boost::get<T>(m_projection);
}

template <typename T>
const T &ProjectionModel::projection() const
{
    return boost::get<T>(m_projection);
}

template <typename T>
Polar<T> project(const ProjectionModel &projectionModel, const Spherical<T> &camPts)
{
    return boost::apply_visitor([&](auto projection) {return project(projection, camPts); }, projectionModel.variant());
}

template <typename T>
Spherical<T> reproject(const ProjectionModel &projectionModel, const Polar<T> &normImgPts)
{
    return boost::apply_visitor([&](auto projection) {return reproject(projection, normImgPts); }, projectionModel.variant());
}

template <typename T>
Point2d<T> project(const ProjectionModel &projectionModel, const Point3d<T> &camPts)
{
    return boost::apply_visitor([&](auto projection) {return project(projection, camPts); }, projectionModel.variant());
}

template <typename T>
Point3d<T> reproject(const ProjectionModel &projectionModel, const Point2d<T> &normImgPts)
{
    return boost::apply_visitor([&](auto projection) {return reproject(projection, normImgPts); }, projectionModel.variant());
}

template <typename T>
bool updateProjectionFromNode(ProjectionModel &projectionModel, const boost::property_tree::ptree &node)
{
    T projection;
    if (!readFromNode(projection, node))
        return false;
    projectionModel.setProjection(projection);
    return true;
}

bool OSP_CALIB_EXPORT readFromNode(ProjectionModel &projectionModel, const boost::property_tree::ptree &node);
void OSP_CALIB_EXPORT writeToNode(const ProjectionModel &projectionModel, boost::property_tree::ptree &node);

}

#endif /* OSP_CALIB_PROJECTIONMODEL_H */
