#ifndef MATH_H
#define MATH_H

#include <cmath>
#include <osp/types/coordinates.h>
#include <osp/types/mathelper.h>

namespace osp {

namespace math {

template <typename T, size_t Size>
T norm(const Vec<T, Size> &vec)
{
    T norm = 0;
    for (size_t i = 0; i < Size; ++i)
        norm += vec[i] * vec[i];
    return std::sqrt(norm);
}

template <typename T, size_t Size>
T norm(const Vec<T, Size> &vec0, const Vec<T, Size> &vec1)
{
    T norm = 0;
    for (size_t i = 0; i < Size; ++i)
    {
        T diff = vec0[i]-vec1[i];
        norm += diff * diff;
    }
    return std::sqrt(norm);
}

template <typename T, size_t Size>
inline T dotProduct(const Vec<T, Size> &vec0, const Vec<T, Size> &vec1)
{
    return vec0[0]*vec1[0] + dotProduct(static_cast<Vec<T, Size-1>>(vec0[1]),
            static_cast<Vec<T, Size-1>>(vec0[1]));
}

template <typename T>
inline T dotProduct(const Vec<T, 1> &vec0, const Vec<T, 1> &vec1)
{
    return vec0[0]*vec1[0];
}

template <typename T, size_t Rows, size_t Cols>
Matrix<T, Rows> matMul(Mat<T, Rows, Cols> mat, Matrix<T, Cols> src)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t xStrideSrc = src.stride(1);

    Matrix<T, Rows> dst(rows, cols);

    if (src.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    size_t xStrideDst = dst.stride(1);

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Rows; ++c)
            {
                ptrDst[c] = dotProduct(static_cast<Vec<T, Cols>>(ptrSrc[c]), static_cast<Vec<T, Cols>>(mat[c * Cols]));
            }

            ptrSrc += xStrideSrc;
            ptrDst += xStrideDst;
        }
    }

    return dst;
}

template <typename T, size_t Channels>
Matrix<T, Channels> multiply(const Matrix<T, Channels> &src, T scalar)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t xStrideSrc = src.stride(1);

    Matrix<T, Channels> dst(rows, cols);

    if (src.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    size_t xStrideDst = dst.stride(1);

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc[c] * scalar;
            }

            ptrSrc += xStrideSrc;
            ptrDst += xStrideDst;
        }
    }

    return dst;
}

template <typename T, size_t Channels>
Matrix<T, Channels> multiply(const Matrix<T, Channels> &src, Vec<T, Channels> vec)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t xStrideSrc = src.stride(1);

    Matrix<T, Channels> dst(rows, cols);

    if (src.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    size_t xStrideDst = dst.stride(1);

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc[c] * vec[c];
            }

            ptrSrc += xStrideSrc;
            ptrDst += xStrideDst;
        }
    }

    return dst;
}

template <typename T1, typename T2, size_t Channels>
Matrix<T1, Channels> multiply(const Matrix<T1, Channels> &src0, const Matrix<T2, 1> &src1)
{
    if (src0.rows() != src1.rows() || src0.cols() != src1.cols())
        throw std::invalid_argument("Matrix size does not match.");

    size_t rows = src0.rows();
    size_t cols = src0.cols();
    size_t xStrideSrc0 = src0.stride(1);
    size_t xStrideSrc1 = src1.stride(1);

    Matrix<T1, Channels> dst(rows, cols);

    if (src0.isContinuous() && src1.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    size_t xStrideDst = dst.stride(1);

    for (size_t y = 0; y < rows; ++y)
    {
        const T1 *ptrSrc0 = src0.ptr(y);
        const T2 *ptrSrc1 = src1.ptr(y);
        T1 *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc0[c] * ptrSrc1[0];
            }

            ptrSrc0 += xStrideSrc0;
            ptrSrc1 += xStrideSrc1;
            ptrDst += xStrideDst;
        }
    }

    return dst;
}

template <typename T1, typename T2, size_t Channels>
typename std::enable_if<Channels != 1, Matrix<T1, Channels>>::type multiply(const Matrix<T1, Channels> &src0, const Matrix<T2, Channels> &src1)
{
    if (src0.rows() != src1.rows() || src0.cols() != src1.cols())
        throw std::invalid_argument("Matrix size does not match.");

    size_t rows = src0.rows();
    size_t cols = src0.cols();
    size_t xStrideSrc0 = src0.stride(1);
    size_t xStrideSrc1 = src1.stride(1);

    Matrix<T1, Channels> dst(rows, cols);

    if (src0.isContinuous() && src1.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }
    size_t xStrideDst = dst.stride(1);

    for (size_t y = 0; y < rows; ++y)
    {
        const T1 *ptrSrc0 = src0.ptr(y);
        const T2 *ptrSrc1 = src1.ptr(y);
        T1 *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc0[c] * ptrSrc1[c];
            }

            ptrSrc0 += xStrideSrc0;
            ptrSrc1 += xStrideSrc1;
            ptrDst += xStrideDst;
        }
    }

    return dst;
}

template <typename T>
Matrix<Spherical<T>> cartesianToSpherical(const Matrix<Point3d<T>> &cartesian)
{
    Matrix<Spherical<T>> spherical;
    processElementwise(cartesian, spherical, [&](const auto &src) {
        return Spherical<T>{src};
    });
    return spherical;
}


template <typename T>
Matrix<T, 3> cartesianToSpherical(Matrix<T, 3> cartesian)
{
    size_t rows = cartesian.rows();
    size_t cols = cartesian.cols();

    Matrix<T, 3> spherical(rows, cols);

    if (cartesian.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    const T *p;
    T *r;
    T sqsum;
    for (size_t i = 0; i < rows; ++i)
    {
        p = cartesian.ptr(i);
        r = spherical.ptr(i);

        for (size_t j = 0; j < cols; ++j)
        {
            sqsum = p[0]*p[0] + p[1]*p[1];
            r[0] = std::atan2(p[1], p[0]);
            r[1] = std::atan2(std::sqrt(sqsum), p[2]);
            r[2] = std::sqrt(sqsum + p[2]*p[2]);

            p += cartesian.stride(1);
            r += spherical.stride(1);
        }
    }

    return spherical;
}

template <typename T>
Matrix<Point3d<T>> sphericalToCartesian(const Matrix<Spherical<T>> &spherical)
{
    Matrix<Point3d<T>> cartesian;
    processElementwise(spherical, cartesian, [&](const auto &src) {
        return Point3d<T>{src};
    });
    return cartesian;
}

template <typename T>
Matrix<T, 3> sphericalToCartesian(const Matrix<T, 3> &spherical)
{
    size_t rows = spherical.rows();
    size_t cols = spherical.cols();

    Matrix<T, 3> cartesian(rows, cols);

    if (spherical.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    const T *p;
    T *r;
    T prod;

    for (size_t i = 0; i < rows; ++i)
    {
        p = spherical.ptr(i);
        r = cartesian.ptr(i);
        for (size_t j = 0; j < cols; ++j)
        {
            prod = p[2] * std::sin(p[1]);
            r[0] = prod * std::cos(p[0]);
            r[1] = prod * std::sin(p[0]);
            r[2] = p[2] * std::cos(p[1]);

            p += spherical.stride(1);
            r += cartesian.stride(1);
        }
    }

    return cartesian;
}

template <typename T>
Matrix<Point2d<T>> polarToCartesian(const Matrix<Polar<T>> &polar)
{
    Matrix<Point2d<T>> cartesian;
    processElementwise(polar, cartesian, [&](const auto &src) {
        return Point2d<T>{src};
    });
    return cartesian;
}

template <typename T>
Matrix<T, 2> polarToCartesian(const Matrix<T, 1> &phi, const Matrix<T, 1> &rho)
{
    if (rho.rows() != phi.rows() || rho.cols() != phi.cols())
        throw std::invalid_argument("Matrix size does not match.");

    size_t rows = rho.rows();
    size_t cols = rho.cols();

    Matrix<T, 2> cartesian(rows, cols);

    if (rho.isContinuous() && phi.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    for (size_t i = 0; i < rows; ++i)
    {
        const T *ptrRho = rho.ptr(i);
        const T *ptrPhi = phi.ptr(i);
        T *ptrCart = cartesian.ptr(i);

        for (size_t j = 0; j < cols; ++j)
        {
            ptrCart[0] = ptrRho[0] * std::cos(ptrPhi[0]);
            ptrCart[1] = ptrRho[0] * std::sin(ptrPhi[0]);

            ptrCart += cartesian.stride(1);
            ptrRho += rho.stride(1);
            ptrPhi += phi.stride(1);
        }
    }

    return cartesian;
}

template <typename T>
Matrix<T, 2> polarToCartesian(const Matrix<T, 2> &pol)
{
    return polarToCartesian(pol.channel(0), pol.channel(1));
}

template <typename T>
Matrix<Polar<T>> cartesianToPolar(const Matrix<Point2d<T>> &cartesian)
{
    Matrix<Polar<T>> polar;
    processElementwise(cartesian, polar, [&](const auto &src) {
        return Polar<T>{src};
    });
    return polar;
}

template <typename T>
Matrix<T, 2> cartesianToPolar(const Matrix<T, 1> &x, const Matrix<T, 1> &y)
{
    if (x.rows() != y.rows() || x.cols() != y.cols())
        throw std::invalid_argument("Matrix size does not match.");

    size_t rows = x.rows();
    size_t cols = x.cols();

    Matrix<T, 2> pol(rows, cols);

    if (x.isContinuous() && x.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    for (size_t i = 0; i < rows; ++i)
    {
        const T *ptrX = x.ptr(i);
        const T *ptrY = y.ptr(i);
        T *ptrPol = pol.ptr(i);

        for (size_t j = 0; j < cols; ++j)
        {
            ptrPol[0] = std::atan2(ptrY[0], ptrX[0]);
            ptrPol[1] = std::sqrt(ptrX[0]*ptrX[0] + ptrY[0]*ptrY[0]);

            ptrPol += pol.stride(1);
            ptrX += x.stride(1);
            ptrY += y.stride(1);
        }
    }

    return pol;
}

template <typename T>
Matrix<T, 2> cartesianToPolar(const Matrix<T, 2> &cart)
{
    return cartesianToPolar(cart.channel(0), cart.channel(1));
}

template <typename T, size_t N>
Matrix<T, 1> polyVal(const Matrix<T, 1> &src, Vec<T, N> coeffs)
{
    size_t rows = src.rows();
    size_t cols = src.cols();

    Matrix<T, 1> dst(rows, cols);
    dst.fill(0);

    if (src.isContinuous() && dst.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);
        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t i = 0; i < N; ++i)
            {
                ptrDst[0] += coeffs[i] * std::pow(ptrSrc[0], i);
            }
            ptrSrc += src.stride(1);
            ptrDst += dst.stride(1);
        }
    }

    return dst;
}

template <typename T, size_t N>
Matrix<T, 1> realRoot(const Matrix<T, 1> &src, Vec<T, N> coeffs)
{
    size_t rows = src.rows();
    size_t cols = src.cols();

    Matrix<T, 1> dst(rows, cols, std::numeric_limits<T>::quiet_NaN());

    if (src.isContinuous() && dst.isContinuous())
    {
        cols *= rows;
        rows = 1;
    }

    cv::Mat coeffsCv(1, N, cv::DataType<T>::type, &coeffs);
    T* ptrFirstCoeff = coeffs.data();
    T firstCoeff = ptrFirstCoeff[0];

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);
        for (size_t x = 0; x < cols; ++x)
        {
            ptrFirstCoeff[0] = firstCoeff - ptrSrc[0];
            cv::Mat roots;
            cv::solvePoly(coeffsCv, roots);

            for (int r = 0; r < roots.rows; ++r)
            {
                T *p = roots.ptr<T>(r);
                if (std::abs(p[1]) < std::numeric_limits<T>::epsilon() && p[0] > 0 && (ptrDst[0] != ptrDst[0] || p[0] < ptrDst[0]))
                {
                    ptrDst[0] = p[0];
                }
            }

            ptrSrc += src.stride(1);
            ptrDst += dst.stride(1);
        }
    }

    return dst;
}

template<typename T>
Vec3<T> rodrigues(const Mat<T, 3, 3> &R)
{
    Vec3<T> r;

    cv::Matx<T, 3, 3> R_cv(R.data());
    cv::Matx<T, 3, 3> U, Vt;
    cv::Vec<double, 3> W;

    cv::SVD::compute(R_cv, W, U, Vt);
    R_cv = U*Vt;

    r = { R_cv(2, 1) - R_cv(1, 2),
          R_cv(0, 2) - R_cv(2, 0),
          R_cv(1, 0) - R_cv(0, 1) };

    T s = std::sqrt((r[0]*r[0] + r[1]*r[1] + r[2]*r[2])*0.25);
    T c = (R_cv(0, 0) + R_cv(1, 1) + R_cv(2, 2) - 1)*0.5;
    c = c > 1. ? 1. : c < -1. ? -1. : c;
    T theta = std::acos(c);

    if( s < 1e-5 )
    {
        T t;

        if( c > 0 )
        {
            r = {0, 0, 0};
        }
        else
        {
            t = (R_cv(0, 0) + 1)*0.5;
            r[0] = std::sqrt(std::max(t,0.));
            t = (R_cv(1, 1) + 1)*0.5;
            r[1] = std::sqrt(std::max(t,0.))*(R_cv(0, 1) < 0 ? -1. : 1.);
            t = (R_cv(2, 2) + 1)*0.5;
            r[2] = std::sqrt(std::max(t,0.))*(R_cv(0, 2) < 0 ? -1. : 1.);
            if( std::fabs(r[0]) < std::fabs(r[1]) && std::fabs(r[0]) < std::fabs(r[2]) && (R_cv(1, 2) > 0) != (r[1]*r[2] > 0) )
                r[2] = -r[2];
            theta /= norm(r);
            r = r * theta;
        }
    }
    else
    {
        double vth = 1/(2*s);
        vth *= theta;
        r = r * vth;
    }

    return r;
}

template<typename T>
Mat<T, 3, 3> rodrigues(Vec3<T> r)
{
    Mat<T, 3, 3> R;

    T theta = norm(r);
    if (theta < std::numeric_limits<T>::epsilon())
        return Mat<T, 3, 3>{ 1, 0, 0,
                             0, 1, 0,
                             0, 0, 1 };

    T c = std::cos(theta);
    T s = std::sin(theta);
    T c1 = 1.-c;
    T iTheta = theta ? 1./theta : 0.;

    r = r * iTheta;

    Mat<T, 3, 3> rrt{ r[0]*r[0], r[0]*r[1], r[0]*r[2],
                      r[0]*r[1], r[1]*r[1], r[1]*r[2],
                      r[0]*r[2], r[1]*r[2], r[2]*r[2]  };
    Mat<T, 3, 3> r_x{     0, -r[2],  r[1],
                       r[2],     0, -r[0],
                      -r[1],  r[0],     0  };
    R = c*Mat<T, 3, 3>::eye() + c1*rrt + s*r_x;

    return R;
}

}

}

#endif // MATH_H
