#ifndef OSP_CV_VPTZ_H
#define OSP_CV_VPTZ_H

#include <osp/calib/camera.h>
#include <osp/exports/cv_export.h>
#include <osp/types/mat.h>

namespace osp {

class OSP_CV_EXPORT VPtz
{
public:
    VPtz();
    ~VPtz() = default;

    /*!
     * \brief Set calibration of source camera
     * \param calibSource Calibration data of source camera
     */
    void setSourceCalib(Camera calibSource);

    /*!
     * \brief Set focal length of target camera
     * \param focalLength focal length in mm
     */
    void setFocalLength(double focalLength);

    /*!
     * \brief Set image resolution of target camera
     * \param width Image width
     * \param height Image height
     */
    void setImageResolution(size_t width, size_t height);

    /*!
     * \brief Set pan, tilt, roll
     * \param panRad Pan angle in radians
     * \param tiltRad Tilt angle in radians
     * \param rollRad Roll angle in radians
     */
    void setOrientation(double panAngleRad, double tiltAngleRad, double rollAngleRad);

    /*!
     * \brief Set pan
     * \param panRad Pan angle in radians
     */
    void setPan(double panAngleRad);

    /*!
     * \brief Set tilt
     * \param tiltRad Tilt angle in radians
     */
    void setTilt(double tiltAngleRad);

    /*!
     * \brief Set roll
     * \param rollRad Roll angle in radians
     */
    void setRoll(double rollAngleRad);

    /*!
     * \brief Get focal length
     * \return focal length
     */
    double focalLength();

    /*!
     * \brief Get pan
     * \return Pan angle in radians
     */
    double pan();

    /*!
     * \brief Get tilt
     * \return Tilt angle in radians
     */
    double tilt();

    /*!
     * \brief Get roll
     * \return Roll angle in radians
     */
    double roll();

    /*!
     * \brief Get calibration of source camera
     * \return Calibration data of source camera
     */
    const Camera &calibSource() const;

    /*!
     * \brief Get calibration of target camera
     * \return Calibration data of target camera
     */
    Camera calibTarget() const;

    /*!
     * \brief Transform the source image
     * \param src Source image
     * \note Implemented for: <uint8_t, 1>, <uint8_t, 3>, <int16_t, 1>, <uint16_t, >, <float, 1>,  <float, 3>, <double, 1>,  <double, 3>, 
     */
    template<typename T, size_t Channels>
    osp::Matrix<T, Channels> OSP_CV_EXPORT transform(const osp::Matrix<T, Channels> &src);

private:
    void resetLut();
    void updateLut();

    Camera m_calibSource;
    double m_focalLength;
    size_t m_imageWidth;
    size_t m_imageHeight;
    double m_panAngleRad;
    double m_tiltAngleRad;
    double m_rollAngleRad;

    osp::Matrix<float, 1> m_mapX;
    osp::Matrix<float, 1> m_mapY;

    std::mutex m_mutexParams;
};

}


#endif
