#ifndef DEVICE_H
#define DEVICE_H

#include <thread>
#include <QFuture>
#include <QMatrix4x4>
#include <QMutex>
#include <QObject>
#include <QReadWriteLock>
#include <QTimer>
#include <QUrl>
#include <QVector3D>
#include <osp/exports/ospqmlplugin_export.h>
#include <osp/sdk/device.h>
#include "frame.h"

class OSP_OSPQMLPLUGIN_EXPORT Device : public QObject
{
    Q_OBJECT
    Q_PROPERTY(QString type READ type NOTIFY typeChanged)
    Q_PROPERTY(QString uri READ uri WRITE setUri NOTIFY uriChanged)
    Q_PROPERTY(bool passwordRequired READ passwordRequired NOTIFY passwordRequiredChanged)
    Q_PROPERTY(bool tlsEnabled READ tlsEnabled WRITE setTlsEnabled NOTIFY tlsEnabledChanged)
    Q_PROPERTY(bool ready READ ready NOTIFY readyChanged)
    Q_PROPERTY(bool running READ running NOTIFY runningChanged)
    Q_PROPERTY(float cpuTemp READ cpuTemp NOTIFY cpuTempChanged)
    Q_PROPERTY(float gpuTemp READ gpuTemp NOTIFY gpuTempChanged)
    Q_PROPERTY(float boardTemp READ boardTemp NOTIFY boardTempChanged)
    Q_PROPERTY(Frame *frame READ frame NOTIFY frameChanged)
    Q_PROPERTY(CaptureMode captureMode READ captureMode WRITE setCaptureMode NOTIFY captureModeChanged)
    Q_PROPERTY(ImageCompression imageCompression READ imageCompression WRITE setImageCompression NOTIFY imageCompressionChanged)
    Q_PROPERTY(StereoConfiguration stereoConfiguration READ stereoConfiguration WRITE setStereoConfiguration NOTIFY stereoConfigurationChanged)
    Q_PROPERTY(StereoMatchingMode stereoMatchingMode READ stereoMatchingMode WRITE setStereoMatchingMode NOTIFY stereoMatchingModeChanged)
    Q_PROPERTY(int stereoMatchingResolutionHorizontal READ stereoMatchingResolutionHorizontal WRITE setStereoMatchingResolutionHorizontal NOTIFY stereoMatchingResolutionHorizontalChanged)
    Q_PROPERTY(int stereoMatchingResolutionVertical READ stereoMatchingResolutionVertical WRITE setStereoMatchingResolutionVertical NOTIFY stereoMatchingResolutionVerticalChanged)
    Q_PROPERTY(float fovH READ fovH WRITE setFovH NOTIFY fovHChanged)
    Q_PROPERTY(float fovV READ fovV WRITE setFovV NOTIFY fovVChanged)
    Q_PROPERTY(float pixPerDegree READ pixPerDegree WRITE setPixPerDegree NOTIFY pixPerDegreeChanged)
    Q_PROPERTY(float pixPerDegreeMin READ pixPerDegreeMin NOTIFY pixPerDegreeMinChanged)
    Q_PROPERTY(float pixPerDegreeMax READ pixPerDegreeMax NOTIFY pixPerDegreeMaxChanged)
    Q_PROPERTY(bool exposureTimeAutoLock READ exposureTimeAutoLock WRITE setExposureTimeAutoLock NOTIFY exposureTimeAutoLockChanged)
    Q_PROPERTY(float exposureTimeMin READ exposureTimeMin WRITE setExposureTimeMin NOTIFY exposureTimeMinChanged)
    Q_PROPERTY(float exposureTimeMax READ exposureTimeMax WRITE setExposureTimeMax NOTIFY exposureTimeMaxChanged)
    Q_PROPERTY(float exposureCompensation READ exposureCompensation WRITE setExposureCompensation NOTIFY exposureCompensationChanged)
    Q_PROPERTY(float gainMin READ gainMin WRITE setGainMin NOTIFY gainMinChanged)
    Q_PROPERTY(float gainMax READ gainMax WRITE setGainMax NOTIFY gainMaxChanged)
    Q_PROPERTY(float digitalGainMin READ digitalGainMin WRITE setDigitalGainMin NOTIFY digitalGainMinChanged)
    Q_PROPERTY(float digitalGainMax READ digitalGainMax WRITE setDigitalGainMax NOTIFY digitalGainMaxChanged)
    Q_PROPERTY(float frameRate READ frameRate WRITE setFrameRate NOTIFY frameRateChanged)
    Q_PROPERTY(QVector3D camPosition READ camPosition NOTIFY camPositionChanged)
    Q_PROPERTY(QVector3D camViewDirection READ camViewDirection NOTIFY camViewDirectionChanged)
    Q_PROPERTY(QVector3D camUpVector READ camUpVector NOTIFY camUpVectorChanged)
    Q_PROPERTY(QMatrix4x4 extrinsic READ extrinsic NOTIFY extrinsicChanged)
    Q_PROPERTY(float fpsViewer READ fpsViewer NOTIFY fpsViewerChanged)
    Q_PROPERTY(float fpsDevice READ fpsDevice NOTIFY fpsDeviceChanged)
public:
    enum CaptureMode
    {
        CaptureMode_RAW,
        CaptureMode_RGBD
    };
    Q_ENUM(CaptureMode)

    enum ImageCompression
    {
        ImageCompression_NONE,
        ImageCompression_LOSSLESS,
        ImageCompression_LOSSY
    };
    Q_ENUM(ImageCompression)

    enum StereoConfiguration
    {
        StereoConfiguration_Binocular,
        StereoConfiguration_Trinocular
    };
    Q_ENUM(StereoConfiguration)

    enum StereoMatchingMode
    {
        StereoMatchingMode_Fast,
        StereoMatchingMode_Quality
    };
    Q_ENUM(StereoMatchingMode)

    Device(QObject *parent = nullptr);
    virtual ~Device();

    QString type();
    QString uri();
    bool passwordRequired();
    bool tlsEnabled();
    bool ready();
    bool running();

    float cpuTemp();
    float gpuTemp();
    float boardTemp();

    Frame *frame();
    CaptureMode captureMode();
    ImageCompression imageCompression();
    bool exposureTimeAutoLock();
    float exposureTimeMin();
    float exposureTimeMax();
    float exposureCompensation();
    float gainMin();
    float gainMax();
    float digitalGainMin();
    float digitalGainMax();
    float frameRate();
    StereoConfiguration stereoConfiguration();
    StereoMatchingMode stereoMatchingMode();
    int stereoMatchingResolutionHorizontal();
    int stereoMatchingResolutionVertical();
    float fovH();
    float fovV();
    float pixPerDegree();
    float pixPerDegreeMin();
    float pixPerDegreeMax();
    float fpsViewer();
    float fpsDevice();

    QMatrix4x4 extrinsic();

    void setType(QString type);
    void setUri(QString uri);
    void setTlsEnabled(bool tlsEnabled);
    void setCaptureMode(CaptureMode mode);
    void setImageCompression(ImageCompression imageCompression);
    void setExposureTimeAutoLock(bool lock);
    void setExposureTimeMin(float min);
    void setExposureTimeMax(float max);
    void setExposureCompensation(float ev);
    void setGainMin(float min);
    void setGainMax(float max);
    void setDigitalGainMin(float min);
    void setDigitalGainMax(float max);
    void setFrameRate(float frameRate);
    void setStereoConfiguration(StereoConfiguration config);
    void setStereoMatchingMode(StereoMatchingMode mode);
    void setStereoMatchingResolutionHorizontal(int resolution);
    void setStereoMatchingResolutionVertical(int resolution);
    void setFovH(float fovH);
    void setFovV(float fovV);
    void setPixPerDegree(float pixPerDegree);

    Q_INVOKABLE bool open();
    Q_INVOKABLE void close();
    Q_INVOKABLE bool login(QString password = QString());
    Q_INVOKABLE bool setPassword(QString password);
    Q_INVOKABLE bool reboot();
    Q_INVOKABLE bool poweroff();
    Q_INVOKABLE bool start();
    Q_INVOKABLE void stop();
    Q_INVOKABLE bool saveCameraParams(QUrl saveDir);

    QVector3D camPosition();
    QVector3D camViewDirection();
    QVector3D camUpVector();

    osp::Device *internalDevice();

    void updateCamExtrinsics();

signals:
    void typeChanged();
    void uriChanged();
    void tlsEnabledChanged();
    void passwordRequiredChanged();
    void readyChanged();
    void runningChanged();
    void cpuTempChanged();
    void gpuTempChanged();
    void boardTempChanged();
    void frameChanged();
    void captureModeChanged();
    void imageCompressionChanged();
    void exposureTimeAutoLockChanged();
    void exposureTimeMinChanged();
    void exposureTimeMaxChanged();
    void exposureCompensationChanged();
    void gainMinChanged();
    void gainMaxChanged();
    void digitalGainMinChanged();
    void digitalGainMaxChanged();
    void frameRateChanged();
    void stereoConfigurationChanged();
    void stereoMatchingModeChanged();
    void stereoMatchingResolutionHorizontalChanged();
    void stereoMatchingResolutionVerticalChanged();
    void fovHChanged();
    void fovVChanged();
    void pixPerDegreeChanged();
    void pixPerDegreeMinChanged();
    void pixPerDegreeMaxChanged();
    void camPositionChanged();
    void camViewDirectionChanged();
    void camUpVectorChanged();
    void extrinsicChanged();
    void fpsViewerChanged();
    void fpsDeviceChanged();

private slots:
    void onReadyChanged();

    void updateFrame(Frame *frame);

    void updateTemperatures();
    void updateCaptureMode();
    void updateImageCompression();
    void updateExposureAutoLock();
    void updateExposureRange();
    void updateExposureCompensation();
    void updateGainRange();
    void updateDigitalGainRange();
    void updateFrameRate();
    void updateStereoConfiguration();
    void updateStereoMatchingMode();
    void updateStereoMatchingResolution();
    void updateTargetCalib();

private:
    void setRunning(bool running);
    void grabbingLoop();

    QString m_type;
    QString m_uri;
    bool m_passwordRequired;
    bool m_tlsEnabled;
    bool m_running;
    Frame *m_frame;

    float m_cpuTemp;
    float m_gpuTemp;
    float m_boardTemp;

    CaptureMode m_captureMode;
    ImageCompression m_imageCompression;
    bool m_exposureAutoLock;
    float m_exposureMin;
    float m_exposureMax;
    float m_exposureCompensation;
    float m_gainMin;
    float m_gainMax;
    float m_digitalGainMin;
    float m_digitalGainMax;
    float m_frameRate;
    StereoConfiguration m_stereoConfiguration;
    StereoMatchingMode m_stereoMatchingMode;
    osp::Size<size_t> m_stereoMatchingResolution;
    osp::Camera m_targetCalib;
    float m_fovH;
    float m_fovV;
    float m_pixPerDegree;
    QMatrix4x4 m_extrinsic;
    std::atomic<float> m_fpsViewer;
    std::atomic<float> m_fpsDevice;

    QVector3D m_camPosition;
    QVector3D m_camViewDirection;
    QVector3D m_camUpVector;

    std::unique_ptr<osp::Device> m_device;

    QReadWriteLock m_deviceMutex;
    QMutex m_frameMutex;

    QTimer m_tempUpdateTimer;
    QFuture<void> m_tempUpdateFuture;

    std::thread m_grabbingThread;
};


#endif // DEVICE_H
