/** \file
 * \brief Device classes
 */

#ifndef OSP_DEVICE_H
#define OSP_DEVICE_H

#include <atomic>
#include <cassert>
#include <chrono>
#include <iostream>
#include <list>
#include <memory>

#include <osp/calib/camera.h>
#include <osp/exports/sdk_export.h>
#include <osp/types/mat.h>
#include <osp/types/range.h>
#include <osp/types/size.h>

namespace osp {

/*!
 * \brief A simple status class.
 */
class OSP_SDK_EXPORT Status
{
public:
    enum class Code
    {
        Success,            /*!< Success */
        Unavailable,        /*!< Ressource unavailable */
        Unauthenticated,    /*!< Authentication failed */
        InvalidArgument,    /*!< Invalid argument was provided */
        PermissionDenied,   /*!< Permission was denied */
        Unknown             /*!< Unknown error */
    };

    Status(Code code, std::string errMsg = std::string());

    /*!
     * \brief Check for status ok
     * \return true if status code is `Success`, false otherwise
     */
    bool ok();

    /*!
     * \brief Return status code
     * \return Status code
     */
    Code statusCode();

    /*!
     * \brief Return error message
     * \return Error message string
     */
    std::string errorMessage();

private:
    Code m_code;
    std::string m_errMsg;
};

/*!
 * \brief Structure that holds the sensor data.
 */
struct Frame
{
    /*!
     * \brief The Metadata struct holds additional information about the Frame
     */
    struct Metadata
    {
        Metadata()
            : timestamp(0)
            , sensorAnalogGain(0)
            , sensorExposureTime(0)
            , sensorSensitivity(0)
            , ispDigitalGain(0)
            , sceneLux(0)
            , achievedFps(0)
        {}

        int64_t timestamp;              /*!< Frame timestamp */
        float sensorAnalogGain;         /*!< Sensor analog gain */
        uint64_t sensorExposureTime;    /*!< Sensor exposure time in ns*/
        uint32_t sensorSensitivity;     /*!< ISO value */
        float ispDigitalGain;           /*!< Digital gain */
        float sceneLux;                 /*!< Estimated scene brightness (in lux) */
        float achievedFps;              /*!< Currently achieved frames per second on the sensor device */
    };

    Metadata metadata;                  /*!< Frame metadata */
    Matrix<uint8_t, 3> sourceImage0;    /*!< Source image from camera 0 */
    Matrix<uint8_t, 3> sourceImage1;    /*!< Source image from camera 1 */
    Matrix<uint8_t, 3> sourceImage2;    /*!< Source image from camera 2 */
    Matrix<uint8_t, 3> image;           /*!< Target (undistorted) image from camera 0 */
    Matrix<uint16_t, 1> distanceMap;    /*!< Distance map */
    Matrix<Point3d<float>> pointcloud;  /*!< Pointcloud */
};

struct PasswordCredentials
{
    std::string password;
};

enum class CameraModelType
{
    Camera0,
    Camera1,
    Camera2,
//    StereoLeftA,
//    StereoRightA,
//    StereoLeftB,
//    StereoRightB,
    Target
};

enum class StereoMatchingMode
{
    Quality,    /*!< Quality stereo matching algorithm */
    Fast        /*!< Fast stereo matching algorithm */
};

enum class StereoMatchingResolution
{
    Low,        /*!< Low stereo matching resolution */
    Middle,     /*!< Middle stereo matching resolution */
    High,       /*!< High stereo matching resolution */
};

enum class TemperatureSensor
{
    CPU,
    GPU,
    Board
};

enum class ImageCompression
{
    None,       /*!< No compression */
    Lossless,   /*!< Lossless compression */
    Lossy       /*!< Lossy compression */
};

/*!
 * \brief The StreamSettings struct defines the settings that are used for
 *        streaming the data from the sensor device.
 */
struct StreamSettings
{
    StreamSettings() : imageCompression{ImageCompression::None} {}

    ImageCompression imageCompression;  /*!< Image compression algorithm */
};

struct StereoMatchingBmParams
{
    // not implemented
};

struct StereoMatchingSgmParams
{
    enum ScanLineDirection
    {
        ScanLineLeftRight           = 1 << 0,   /*!< Aggregate cost from left to right horizontally */
        ScanLineTopLeftBottomRight  = 1 << 1,   /*!< Aggregate cost from left to right diagonally starting from the top */
        ScanLineTopBottom           = 1 << 2,   /*!< Aggregate cost from top to bottom vertically */
        ScanLineTopRightBottomLeft  = 1 << 3,   /*!< Aggregate cost from right to left diagonally starting from the top */
        ScanLineRightLeft           = 1 << 4,   /*!< Aggregate cost from right to left horizontally */
        ScanLineBottomRightTopLeft  = 1 << 5,   /*!< Aggregate cost from right to left diagonally starting from the bottom */
        ScanLineBottomTop           = 1 << 6,   /*!< Aggregate cost from bottom to top vertically */
        ScanLineBottomLeftTopRight  = 1 << 7,   /*!< Aggregate cost from left to right diagonally starting from the bottom */

        /*!< Aggregate cost from four directions forming a cross */
        ScanLineCross = ScanLineLeftRight | ScanLineRightLeft | ScanLineTopBottom | ScanLineBottomTop,

        ScanLineAll                 = 0xFF      /*!< Aggregate cost over all scan lines */
    };

    StereoMatchingSgmParams()
        : maxDisparity(96)
        , p1(8)
        , p2(32)
        , sad(3)
        , ctWinSize(3)
        , hcWinSize(5)
        , clip(31)
        , maxDiff(5)
        , uniqueness(15)
        , scanlines(ScanLineAll)
        , filterTopArea(false)
        , pyramidalStereo(true)
    {}

    int32_t maxDisparity;           /*!< The maximum disparity value */
    int32_t p1;                     /*!< Smoothness penalty */
    int32_t p2;                     /*!< Discontiguous penalty */
    int32_t sad;                    /*!< Specifies the average window size for sum of absolute differences (set to 1 for original SGM) */
    int32_t ctWinSize;              /*!< Specifies the census transform window size */
    int32_t hcWinSize;              /*!< Specifies the hamming cost window size */
    int32_t clip;                   /*!< Clip value for cost (used to preserve overflow in cost function) */
    int32_t maxDiff;                /*!< Maximum allowed difference (in integer pixel units) in the left-right disparity check */
    int32_t uniqueness;             /*!< Margin in percentage by which the best (minimum) computed cost function value
                                         should “win” the second best value to consider the found match correct */
    int scanlines;                  /*!< Scan line directions mask (Any bit-OR of ScanLineDirection values is suitable)*/
    bool filterTopArea;             /*!< Filter cost at top image area with low gradients */
    bool pyramidalStereo;           /*!< Use pyramidal scheme: lower resolution imagery for nearby objects and
                                         the full resolution for far-away objects */
};

/*!
 * \brief The StereoMatchingParams struct contains all parameters used by the
 *        stereo algorithms.
 */
struct StereoMatchingParams
{
    StereoMatchingBmParams bmParams;
    StereoMatchingSgmParams sgmParams;
};


/*!
 * \brief Comparison operators
 */
OSP_SDK_EXPORT bool operator==(const StereoMatchingBmParams &lhs, const StereoMatchingBmParams &rhs);
OSP_SDK_EXPORT bool operator==(const StereoMatchingSgmParams &lhs, const StereoMatchingSgmParams &rhs);
OSP_SDK_EXPORT bool operator==(const StereoMatchingParams &lhs, const StereoMatchingParams &rhs);

/*!
 * \brief The Device class controls a single sensor device.
 */
class OSP_SDK_EXPORT Device
{
public:
    class Private;

    /*!
     * \brief The Info struct holds some information about a device
     */
    struct Info
    {
        std::string type;           /*!< Device type */
        std::string serialNumber;   /*!< Serial number of the device */
        std::string uri;            /*!< Device address */
        bool passwordRequired;      /*!< Requires login using a password */
    };

    /*!
     * \brief The capture modes
     */
    enum class CaptureMode
    {
        RAW,    /*!< Capture source images [0..2] */
        RGBD    /*!< Capture undistorted image from camera 0 and corresponding distance map */
    };

    /*!
     * \brief The stereo configuration modes
     */
    enum class StereoConfiguration
    {
        Binocular,
        Trinocular
    };

    /*!
     * \brief Destructor
     */
    virtual ~Device();

    /*!
     * \brief Open device
     * \param uri Device address
     * \return Status
     */
    Status open(std::string uri);

    /*!
     * \brief close Device
     */
    void close();

    /*!
     * \brief Log in to device using password credentials
     * \param credentials PasswordCredentials object containing the password
     * \return Status
     */
    Status login(const PasswordCredentials &credentials);

    /*!
     * \brief Log out from device
     */
    void logout();

    /*!
     * \brief Set the password credentials for accessing the device
     * \param credentials PasswordCredentials object containing the new password
     * \return true on success, false otherwise
     */
    bool setPasswordCredentials(const PasswordCredentials &credentials);

    /*!
     * \brief Get device infomation
     * \return Info about connected device
     */
    Info info();

    /*!
     * \brief Reboot the device
     * \return true on success, false otherwise
     */
    bool reboot();

    /*!
     * \brief Shutdown the device
     * \return true on success, false otherwise
     */
    bool poweroff();

    /*!
     * \brief Start processing of current mode
     * \return Status
     */
    Status start();

    /*!
     * \brief Stop processing
     */
    void stop();

    /*!
     * \brief Check device running
     * \return true if running, false otherwise
     */
    bool running();

    /*!
     * \brief Get a frame from the sensor
     * \param timeout Timeout for receiving the frame. If timeout expires, the return value will be empty.
     * \return true if running, false otherwise
     */
    std::unique_ptr<osp::Frame> getFrame(std::chrono::milliseconds timeout = std::chrono::milliseconds(0));

    /*!
     * \brief Get temperature of the CPU
     * \param sensor temperature sensor
     * \return temperature in °C
     */
    float temperature(TemperatureSensor sensor);

    /*!
     * \brief Request camera model params
     * \param type Type of the camera model to be returned
     * \return Camera model
     */
    Camera getCameraModel(CameraModelType type);

    /*!
     * \brief Set camera model params
     * \param type Type of the camera model to be set
     * \param model
     */
    void setCameraModel(CameraModelType type, Camera model);

    /*!
     * \brief Set current camera model params as default
     */
    void saveCurrentCameraModelsAsDefault();

    /*!
     * \brief Set extrinsic camera parameters
     * \param extrinsic Extrinsic parameters
     */
    void setExtrinsic(Extrinsic extrinsic);

    /*!
     * \brief Request extrinsic camera parameters
     * \return Extrinsic parameters
     */
    Extrinsic getExtrinsic();

    /*!
     * \brief Get current capture mode
     * \return Current capture mode (RGBD/RAW)
     */
    CaptureMode captureMode() const;

    /*!
     * \brief Set capture mode
     * \param mode Capture mode
     * \return Status
     */
    Status setCaptureMode(CaptureMode mode);

    /*!
     * \brief Check if auto exposure is locked
     * \return true if locked, false otherwise
     */
    bool exposureAutoLock();

    /*!
     * \brief Return exposure time range
     * \return Range in [ns]
     */
    Range<uint64_t> exposureRange();

    /*!
     * \brief Return exposure compensation
     * \return exposure compensation in ev
     */
    float exposureCompensation();

    /*!
     * \brief Return analog gain range
     * \return Range
     */
    Range<float> gainRange();

    /*!
     * \brief Return digital gain range
     * \return Range
     */
    Range<float> digitalGainRange();

    /*!
     * \brief Return capture frame rate
     * \return frames per second
     */
    float frameRate();

    /*!
     * \brief Lock/Unlock current exposure time
     * \param lock If true, the current exposure time will be locked,
     * otherwise the auto controls can change it.
     * \return true on success, false otherwise
     */
    bool setExposureAutoLock(bool lock);

    /*!
     * \brief Set exposure time range
     * \param range exposure time range in ns [30000...400000000]
     * \note The maximum reachable exposure time depends on the selected
     * frame rate: \f$ \textrm{max\_exp\_ns} = 1e9/\textrm{fps} \f$
     * Therefore you should reduce the frame rate to increase the maximum
     * exposure time.
     * \return true on success, false otherwise
     */
    bool setExposureRange(Range<uint64_t> range);

    /*!
     * \brief Set exposure compensation
     * \param ev exposure compensation in ev [-10...10]
     * \return true on success, false otherwise
     */
    bool setExposureCompensation(float ev);

    /*!
     * \brief Set analog gain range
     * \param range analog gain range [1...251]
     * \return true on success, false otherwise
     */
    bool setGainRange(Range<float> range);

    /*!
     * \brief Set digital gain range
     * \param range digital gain range [1...256]
     * \return true on success, false otherwise
     */
    bool setDigitalGainRange(Range<float> range);

    /*!
     * \brief Set capture frame rate
     * \param frameRate frames per second [1...30]
     * \return true on success, false otherwise
     */
    bool setFrameRate(float frameRate);

    /*!
     * \brief Get current stereo configuration
     * \return stereo configuration
     */
    StereoConfiguration stereoConfiguration();

    /*!
     * \brief Set stereo configuration
     * \param config stereo configuration
     * \return true on success, false otherwise
     */
    bool setStereoConfiguration(StereoConfiguration config);

    /*!
     * \brief Get current stereo matching mode
     * \return stereo matching mode
     */
    StereoMatchingMode stereoMatchingMode();

    /*!
     * \brief Set stereo matching mode
     * \param mode stereo matching mode
     * \return true on success, false otherwise
     */
    bool setStereoMatchingMode(StereoMatchingMode mode);

    /*!
     * \brief Get current stereo matching resolution
     * \return stereo matching resolution
     */
    osp::Size<size_t> stereoMatchingResolution();

    /*!
     * \brief Set stereo matching resolution
     * \param resolution stereo matching resolution
     * \return true on success, false otherwise
     */
    bool setStereoMatchingResolution(const osp::Size<size_t> &resolution);

    /*!
     * \brief Get current settings for streaming between device and client
     * \return Stream settings object
     */
    StreamSettings streamSettings() const;

    /*!
     * \brief Set settings for streaming between device and client
     * \param Stream settings object
     */
    void setStreamSettings(const StreamSettings &streamSettings);


    /*!
     * \brief Get the current stereo matching parameters
     * \return Stereo matching parameters
     */
    StereoMatchingParams stereoMatchingParams() const;

    /*!
     * \brief Set the parameters for stereo matching
     * \param Stereo matching parameters
     */
    bool setStereoMatchingParams(const StereoMatchingParams &stereoMatchingParams);


protected:
    Device();

private:
    std::unique_ptr<Private> m_p;
};

/*!
 * \brief The DeviceManager class provides device discovery and opens
 * new connections to devices.
 */
class OSP_SDK_EXPORT DeviceManager
{
public:
    class Private;
    virtual ~DeviceManager();

    /*!
     * \brief Get device manager
     * \return Device manager singleton
     */
    static DeviceManager &get();

    /*!
     * \brief Return discovered devices
     * \return Vector of device information
     */
    std::vector<Device::Info> getDevices();

    /*!
     * \brief Request device
     * \param info Device information object
     * \param enable TLS encryption
     * \return Unique Pointer to Device. Empty, if the device could not be opened.
     */
    std::unique_ptr<Device> request(const Device::Info &info, bool useTls = false);

    /*!
     * \brief Request device
     * \param uri Device address
     * \param enable TLS encryption
     * \return Unique Pointer to Device. Empty, if the device could not be opened.
     */
    std::unique_ptr<Device> request(const std::string &uri, bool useTls = false);

    /*!
     * \brief Initiate a rescan for available devices.
     */
    void rescan();

protected:
    DeviceManager();

private:
    std::unique_ptr<Private> m_p;
};

}

#endif // OSP_DEVICE_H
