#ifndef OSP_CALIB_COORDINATES_H
#define OSP_CALIB_COORDINATES_H

#include <cmath>
#include <ostream>

namespace osp {

template <typename T> struct Polar;
template <typename T> struct Spherical;
template <typename T> struct Point2d;
template <typename T> struct Point3d;

template <typename T>
struct Polar
{
    using ValueType = T;

    Polar(T _phi = 0, T _rho = 0);
    template<typename U>
    Polar(const Polar<U> &);
    explicit Polar(const Point2d<T> &);

    T phi;
    T rho;
};

template <typename T>
struct Spherical
{
    using ValueType = T;

    Spherical(T _phi = 0, T _theta = 0, T _rho = 0);
    template<typename U>
    Spherical(const Spherical<U> &);
    explicit Spherical(const Point3d<T> &);

    T phi;
    T theta;
    T rho;
};

template <typename T>
struct Point2d
{
    using ValueType = T;

    Point2d(T _x = 0, T _y = 0);
    template<typename U>
    Point2d(const Point2d<U> &);
    explicit Point2d(const Polar<T> &);

    T x;
    T y;
};

template <typename T>
struct Point3d
{
    using ValueType = T;

    Point3d(T _x = 0, T _y = 0, T _z = 0);
    template<typename U>
    Point3d(const Point3d<U> &);
    explicit Point3d(const Spherical<T> &);

    T x;
    T y;
    T z;
};

template <typename T>
Polar<T>::Polar(T _phi, T _rho)
    : phi{_phi}
    , rho{_rho}
{}

template <typename T>
template <typename U>
Polar<T>::Polar(const Polar<U> &other)
    : phi{static_cast<T>(other.phi)}
    , rho{static_cast<T>(other.rho)}
{}

template <typename T>
Polar<T>::Polar(const Point2d<T> &cart)
    : phi{std::atan2(cart.y, cart.x)}
    , rho{std::sqrt(cart.x*cart.x + cart.y*cart.y)}
{}

template <typename T>
Spherical<T>::Spherical(T _phi, T _theta, T _rho)
    : phi{_phi}
    , theta{_theta}
    , rho{_rho}
{}

template <typename T>
template <typename U>
Spherical<T>::Spherical(const Spherical<U> &other)
    : phi{static_cast<T>(other.phi)}
    , theta{static_cast<T>(other.theta)}
    , rho{ static_cast<T>(other.rho)}
{}

template <typename T>
Spherical<T>::Spherical(const Point3d<T> &cart)
{
    auto sqsum = cart.x*cart.x + cart.y*cart.y;
    phi = std::atan2(cart.y, cart.x);
    theta = std::atan2(std::sqrt(sqsum), cart.z);
    rho = std::sqrt(sqsum + cart.z*cart.z);
}

template <typename T>
Point2d<T>::Point2d(T _x, T _y)
    : x{_x}
    , y{_y}
{}

template <typename T>
template <typename U>
Point2d<T>::Point2d(const Point2d<U> &other)
    : x{static_cast<T>(other.x)}
    , y{static_cast<T>(other.y)}
{}

template <typename T>
Point2d<T>::Point2d(const Polar<T> &polar)
    : x{polar.rho * std::cos(polar.phi)}
    , y{polar.rho * std::sin(polar.phi)}
{}

template <typename T>
Point3d<T>::Point3d(T _x, T _y, T _z)
    : x{_x}
    , y{_y}
    , z{_z}
{}

template <typename T>
template <typename U>
Point3d<T>::Point3d(const Point3d<U> &other)
    : x{static_cast<T>(other.x)}
    , y{static_cast<T>(other.y)}
    , z{static_cast<T>(other.z)}
{}

template <typename T>
Point3d<T>::Point3d(const Spherical<T> &spherical)
{
    auto prod = spherical.rho * std::sin(spherical.theta);
    x = prod * std::cos(spherical.phi);
    y = prod * std::sin(spherical.phi);
    z = spherical.rho * std::cos(spherical.theta);
}

template <typename T>
Point2d<T> operator+ (const Point2d<T> &lhs, const Point2d<T> &rhs)
{
    return {
        lhs.x + rhs.x,
        lhs.y + rhs.y
    };
}

template <typename T>
Point2d<T> operator- (const Point2d<T> &lhs, const Point2d<T> &rhs)
{
    return {
        lhs.x - rhs.x,
        lhs.y - rhs.y
    };
}

template <typename T>
Point3d<T> operator+ (const Point3d<T> &lhs, const Point3d<T> &rhs)
{
    return {
        lhs.x + rhs.x,
        lhs.y + rhs.y,
        lhs.z + rhs.z,
    };
}

template <typename T>
Point3d<T> operator- (const Point3d<T> &lhs, const Point3d<T> &rhs)
{
    return {
        lhs.x - rhs.x,
        lhs.y - rhs.y,
        lhs.z - rhs.z,
    };
}


template <typename T>
Polar<T> operator* (float factor, const Polar<T>& coordinate)
{
    return {
        coordinate.phi,
        factor*coordinate.rho
    };
}

template <typename T>
Polar<T> operator* (const Polar<T>& coordinate, float factor)
{
    return factor * coordinate;
}

template <typename T>
Spherical<T> operator* (float factor, const Spherical<T>& coordinate)
{
    return {
        coordinate.phi,
        coordinate.theta,
        factor*coordinate.rho
    };
}

template <typename T>
Spherical<T> operator* (const Spherical<T>& coordinate, float factor)
{
    return factor * coordinate;
}

template <typename T>
Point2d<T> operator* (float factor, const Point2d<T>& coordinate)
{
    return {
        coordinate.x*factor,
        coordinate.y*factor
    };
}

template <typename T>
Point2d<T> operator* (const Point2d<T>& coordinate, float factor)
{
    return factor * coordinate;
}

template <typename T>
Point3d<T> operator* (float factor, const Point3d<T>& coordinate)
{
    return {
        coordinate.x*factor,
        coordinate.y*factor,
        coordinate.z*factor
    };
}

template <typename T>
Point3d<T> operator* (const Point3d<T>& coordinate, float factor)
{
    return factor * coordinate;
}

template <typename T>
std::ostream& operator<<(std::ostream& stream, const Point2d<T>& cart)
{
    stream << "{ ";
    stream << cart.x;
    stream << ", " << cart.y;
    stream << " }";
    return stream;
}

template <typename T>
std::ostream& operator<<(std::ostream& stream, const Point3d<T>& cart)
{
    stream << "{ ";
    stream << cart.x;
    stream << ", " << cart.y;
    stream << ", " << cart.z;
    stream << " }";
    return stream;
}

template <typename T>
std::ostream& operator<<(std::ostream& stream, const Polar<T>& pol)
{
    stream << "{ ";
    stream << pol.phi;
    stream << ", " << pol.rho;
    stream << " }";
    return stream;
}

template <typename T>
std::ostream& operator<<(std::ostream& stream, const Spherical<T>& sph)
{
    stream << "{ ";
    stream << sph.phi;
    stream << ", " << sph.theta;
    stream << ", " << sph.rho;
    stream << " }";
    return stream;
}

template <typename T> struct Projected {};

template <typename T>
struct Projected<Point3d<T>>
{
    using type = Point2d<T>;
};

template <typename T>
struct Projected<Spherical<T>>
{
    using type = Polar<T>;
};

template <typename T> struct Reprojected {};

template <typename T>
struct Reprojected<Point2d<T>>
{
    using type = Point3d<T>;
};

template <typename T>
struct Reprojected<Polar<T>>
{
    using type = Spherical<T>;
};

}

#endif
