#ifndef OSP_MAT_DETAILS_H
#define OSP_MAT_DETAILS_H

namespace osp {

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols>::Mat()
{
    m_data.fill(0);
}

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols>::Mat(std::initializer_list<T> initializer)
{
    std::copy(initializer.begin(), initializer.end(), m_data.begin());
}

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols>::Mat(const Mat &other)
    : m_data(other.m_data)
{
}

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> Mat<T, Rows, Cols>::eye()
{
    static_assert(Rows == Cols, "Matrix must be square.");

    Mat mat;
    for (size_t i = 0; i < Rows; ++i)
        mat.at(i, i) = 1;
    return mat;
}

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> &Mat<T, Rows, Cols>::operator=(Mat<T, Rows, Cols> other)
{
    using std::swap;
    swap(other, *this);
    return *this;
}

template <typename T, size_t Rows, size_t Cols>
Mat<T, Rows, Cols>::Mat(const T *data)
{
    std::copy(data, data+Rows*Cols, m_data.begin());
}

template <typename T, size_t Rows, size_t Cols>
size_t Mat<T, Rows, Cols>::rows() const
{
    return Rows;
}

template <typename T, size_t Rows, size_t Cols>
size_t Mat<T, Rows, Cols>::cols() const
{
    return Cols;
}

template <typename T, size_t Rows, size_t Cols>
void Mat<T, Rows, Cols>::fill(const T &val)
{
    m_data.fill(val);
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::reference &Mat<T, Rows, Cols>::at(size_t y, size_t x)
{
    return m_data.at(y*Cols + x);
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::const_reference &Mat<T, Rows, Cols>::at(size_t y, size_t x) const
{
    return m_data.at(y*Cols + x);
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::reference Mat<T, Rows, Cols>::operator[](size_t idx)
{
    return m_data[idx];
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::const_reference Mat<T, Rows, Cols>::operator[](size_t idx) const
{
    return m_data[idx];
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::pointer Mat<T, Rows, Cols>::data()
{
    return m_data.data();
}

template <typename T, size_t Rows, size_t Cols>
typename std::array<T, Rows*Cols>::const_pointer Mat<T, Rows, Cols>::data() const
{
    return m_data.data();
}

template <typename SrcType, typename DstType, size_t Rows, size_t Cols, typename Func, typename ...Args>
void processElementwise(const Mat<SrcType, Rows, Cols> &src, Mat<DstType, Rows, Cols> &dst, Func f, Args&&... args)
{
    constexpr size_t N = Rows*Cols;
    for (size_t i = 0; i < N; ++i)
        dst[i] = f(src[i], std::forward<Args>(args)...);
}


template <typename SrcType0, typename SrcType1, typename DstType, size_t Rows, size_t Cols, typename Func, typename ...Args>
void processElementwise(const Mat<SrcType0, Rows, Cols> &src0, const Mat<SrcType1, Rows, Cols> &src1, Mat<DstType, Rows, Cols> &dst, Func f, Args&&... args)
{
    constexpr size_t N = Rows*Cols;
    for (size_t i = 0; i < N; ++i)
        dst[i] = f(src0[i], src1[i], std::forward<Args>(args)...);
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator*(const Mat<T, Rows, Cols> &src, const U &val)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src, dst, [&val](const T &srcVal){ return srcVal * val; });
    return dst;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const Mat<T, Rows, Cols> &src, const U &val)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src, dst, [&val](const T &srcVal){ return srcVal + val; });
    return dst;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator-(const Mat<T, Rows, Cols> &src, const U &val)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src, dst, [&val](const T &srcVal){ return srcVal - val; });
    return dst;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator/(const Mat<T, Rows, Cols> &src, const U &val)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src, dst, [&val](const T &srcVal){ return srcVal / val; });
    return dst;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator*(const U &val, const Mat<T, Rows, Cols> &src)
{
    return src*val;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const U &val, const Mat<T, Rows, Cols> &src)
{
    return src+val;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const Mat<T, Rows, Cols> &src0, const Mat<U, Rows, Cols> &src1)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src0, src1, dst, [](const T &src0Val, const T &src1Val){ return src0Val + src1Val; });
    return dst;
}

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator-(const Mat<T, Rows, Cols> &src0, const Mat<U, Rows, Cols> &src1)
{
    Mat<T, Rows, Cols> dst;
    processElementwise(src0, src1, dst, [](const T &src0Val, const T &src1Val){ return src0Val - src1Val; });
    return dst;
}

template <size_t N, class T>
Vec<T ,N> makeVec(const T &v) {
    Vec<T, N> ret;
    ret.fill(v);
    return ret;
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix()
    : m_rows(0)
    , m_cols(0)
    , m_stride{0, 0}
    , m_offset(0)
    , m_continuous(true)
    , m_buf(0)
{
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(size_t rows, size_t cols, T val)
    : m_rows(rows)
    , m_cols(cols)
    , m_stride{cols*Channels, Channels}
    , m_offset(0)
    , m_continuous(true)
    , m_buf(rows*cols*Channels)
{
    fill(val);
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(size_t rows, size_t cols, Vec<T, Channels> val)
    : m_rows(rows)
    , m_cols(cols)
    , m_stride{cols*Channels, Channels}
    , m_offset(0)
    , m_continuous(true)
    , m_buf(rows*cols*Channels)
{
    fill(val);
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(size_t rows, size_t cols, T *data, Vec<size_t, 2> stride, size_t offset)
    : m_rows(rows)
    , m_cols(cols)
    , m_stride(stride)
    , m_offset(offset)
    , m_continuous(((stride[0] == m_cols*Channels) && (stride[1] == Channels)) ? true : false)
{
    if (stride[0] == 0 || stride[1] == 0)
    {
        m_stride[0] = cols*Channels;
        m_stride[1] = Channels;
        m_continuous = true;

    }
    m_buf = Buffer<T>(m_rows * m_stride[0], data);
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(size_t rows, size_t cols, Buffer<T> buf, Vec<size_t, 2> stride, size_t offset)
    : m_rows(rows)
    , m_cols(cols)
    , m_stride(stride)
    , m_offset(offset)
    , m_continuous(((stride[0] == m_cols*Channels) && (stride[1] == Channels)) ? true : false)
    , m_buf(buf)
{
    if (stride[0] == 0 || stride[1] == 0)
    {
        m_stride[0] = cols*Channels;
        m_stride[1] = Channels;
        m_continuous = true;
    }
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(const Matrix<T, Channels> & other)
    : m_rows(other.m_rows)
    , m_cols(other.m_cols)
    , m_stride(other.m_stride)
    , m_offset(other.m_offset)
    , m_continuous(other.m_continuous)
    , m_buf(other.m_buf)
{
}

template<typename T, size_t Channels>
Matrix<T, Channels>::Matrix(Matrix<T, Channels> && other)
    : m_rows(std::move(other.m_rows))
    , m_cols(std::move(other.m_cols))
    , m_stride(std::move(other.m_stride))
    , m_offset(std::move(other.m_offset))
    , m_continuous(std::move(other.m_continuous))
    , m_buf(std::move(other.m_buf))
{
}

template<typename T, size_t Channels>
Matrix<T, Channels> & Matrix<T, Channels>::operator=(Matrix<T, Channels> other)
{
    other.swap(*this);
    return *this;
}

template<typename T, size_t Channels>
void Matrix<T, Channels>::swap(Matrix<T, Channels> & other) noexcept
{
    std::swap(this->m_rows, other.m_rows);
    std::swap(this->m_cols, other.m_cols);
    std::swap(this->m_stride, other.m_stride);
    std::swap(this->m_offset, other.m_offset);
    std::swap(this->m_continuous, other.m_continuous);
    std::swap(this->m_buf, other.m_buf);
}

template<typename T, size_t Channels>
void Matrix<T, Channels>::resize(size_t rows, size_t cols)
{
    if (this->rows() == rows && this->cols() == cols)
        return;

    *this = Matrix<T, Channels>{rows, cols};
}

template<typename T, size_t Channels>
template <typename DstType>
Matrix<DstType, Channels> Matrix<T, Channels>::convert() const
{
    Matrix<DstType, Channels> matConverted(m_rows, m_cols);
    for (size_t y = 0; y < m_rows; ++y)
    {
        const T *ptrSrc = ptr(y);
        DstType *ptrDst = matConverted.ptr(y);

        for (size_t x = 0; x < m_cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = DstType(ptrSrc[c]);
            }

            ptrSrc += m_stride[1];
            ptrDst += matConverted.stride(1);
        }
    }
    return matConverted;
}

template<typename T, size_t Channels>
Matrix<T, Channels> Matrix<T, Channels>::transpose() const
{
    Matrix<T, Channels> matTransposed(m_cols, m_rows);

    for (size_t y = 0; y < m_rows; ++y)
    {
        const T *ptrSrc = ptr(y);
        T *ptrDst = matTransposed.ptr(0, y);

        for (size_t x = 0; x < m_cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc[c];
            }

            ptrSrc += m_stride[1];
            ptrDst += matTransposed.stride(0);
        }
    }
    return matTransposed;
}

template<typename T, size_t Channels>
T& Matrix<T, Channels>::at(size_t y, size_t x, size_t channel)
{
    return *ptr(y, x, channel);
}

template<typename T, size_t Channels>
const T& Matrix<T, Channels>::at(size_t y, size_t x, size_t channel) const
{
    return *ptr(y, x, channel);
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::rows() const
{
    return m_rows;
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::cols() const
{
    return m_cols;
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::channels() const
{
    return Channels;
}

template<typename T, size_t Channels>
Vec<size_t, 2> Matrix<T, Channels>::stride() const
{
    return m_stride;
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::stride(size_t idx) const
{
    return m_stride[idx];
}

template<typename T, size_t Channels>
bool Matrix<T, Channels>::empty() const
{
    return m_buf.empty();
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::total() const
{
    return m_buf.size();
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::elemSize() const
{
    return sizeof(T);
}

template<typename T, size_t Channels>
size_t Matrix<T, Channels>::bytes() const
{
    return total() * elemSize();
}

template<typename T, size_t Channels>
Buffer<T> Matrix<T, Channels>::internalBuffer() const
{
    return m_buf;
}

template<typename T, size_t Channels>
T* Matrix<T, Channels>::ptr(size_t y) const
{
    size_t idx = y * m_stride[0] + m_offset;

    if (idx > m_buf.size())
        throw std::out_of_range("Index exceeds Matrix dimensions.");

    return m_buf.data() + idx;
}

template<typename T, size_t Channels>
T* Matrix<T, Channels>::ptr(size_t y, size_t x) const
{
    size_t idx = y * m_stride[0] + x * m_stride[1] + m_offset;

    if (idx > m_buf.size())
        throw std::out_of_range("Index exceeds Matrix dimensions.");

    return m_buf.data() + idx;
}

template<typename T, size_t Channels>
T* Matrix<T, Channels>::ptr(size_t y, size_t x, size_t channel) const
{
    size_t idx = y * m_stride[0] + x * m_stride[1] + channel + m_offset;

    if (idx > m_buf.size())
        throw std::out_of_range("Index exceeds Matrix dimensions.");

    return m_buf.data() + idx;
}

template<typename T, size_t Channels>
bool Matrix<T, Channels>::isContinuous() const
{
    return m_continuous;
}

template<typename T, size_t Channels>
Matrix<T, Channels> &Matrix<T, Channels>::makeContinuous()
{
     *this = clone();
     return *this;
}

template<typename T, size_t Channels>
Matrix<T, Channels> Matrix<T, Channels>::clone() const
{
    Matrix matCopy(m_rows, m_cols);
    if (m_continuous)
    {
        std::memcpy(matCopy.m_buf.data(), m_buf.data() + m_offset, m_rows*m_cols*Channels*sizeof(T));
    }
    else
    {
        for (size_t y = 0; y < m_rows; ++y)
        {
            const T *ptrSrc = ptr(y);
            T *ptrDst = matCopy.ptr(y);

            for (size_t x = 0; x < m_cols; ++x)
            {
                for (size_t c = 0; c < Channels; ++c)
                {
                    ptrDst[c] = ptrSrc[c];
                }

                ptrSrc += m_stride[1];
                ptrDst += matCopy.m_stride[1];
            }
        }
    }
    return matCopy;
}

template<typename T, size_t Channels>
void Matrix<T, Channels>::fill(T val)
{
    Vec<T, Channels> vec;
    vec.fill(val);
    fill(vec);
}

template<typename T, size_t Channels>
void Matrix<T, Channels>::fill(Vec<T, Channels> val)
{
    for (size_t y = 0; y < m_rows; ++y)
    {
        T *p = ptr(y);

        for (size_t x = 0; x < m_cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                p[c] = val[c];
            }

            p += m_stride[1];
        }
    }
}

template<typename T, size_t Channels>
template <typename U>
void Matrix<T, Channels>::setTo(T val, Matrix<U, 1> mask)
{
    Vec<T, Channels> vec;
    vec.fill(val);
    setTo(vec, mask);
}

template<typename T, size_t Channels>
template <typename U>
void Matrix<T, Channels>::setTo(Vec<T, Channels> val, Matrix<U, 1> mask)
{
    if (mask.rows() != m_rows || mask.cols() != m_cols)
        throw std::out_of_range("Mask size must match matrix size.");

    size_t maskStrideX = mask.stride(1);

    for (size_t y = 0; y < m_rows; ++y)
    {
        T *p = ptr(y);
        U *pMask = mask.ptr(y);

        for (size_t x = 0; x < m_cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                p[c] = (*pMask) ? val[c] : p[c];
            }

            p += m_stride[1];
            pMask += maskStrideX;
        }
    }
}

template<typename T, size_t Channels>
Matrix<T, 1> Matrix<T, Channels>::channel(size_t c) const
{
    if (c >= Channels)
        throw std::out_of_range("Index exceeds Matrix dimensions.");

    Matrix<T, 1> matChannel(m_rows, m_cols, m_buf, m_stride, m_offset + c);
    return matChannel;
}


template<typename T, size_t Rows, size_t Cols>
inline std::ostream &operator <<(std::ostream &stream, const Mat<T, Rows, Cols> &matrix)
{
    stream << "[";

    for (size_t y = 0; y < Rows; ++y)
    {
        for (size_t x = 0; x < Cols; ++x)
        {
            stream << "\t" << matrix.at(y, x);
        }

        stream << std::endl;
    }
    stream << "]" << std::endl;

    return stream;
}

template<typename T, size_t Channels>
inline std::ostream &operator <<(std::ostream &stream, const Matrix<T, Channels> &matrix)
{
    stream << "----------------------";
    stream << "Size: ";
    stream << matrix.rows();
    stream << "x" << matrix.cols();
    stream << "x" << Channels;
    stream << "----------------------";

    for (size_t c = 0; c < Channels; ++c)
    {
        stream << std::endl;
        stream << "Channel " << c << std::endl;
        stream << "[";

        for (size_t y = 0; y < matrix.rows(); ++y)
        {
            T *p = matrix.ptr(y) + c;

            for (size_t x = 0; x < matrix.cols(); ++x)
            {
                stream << "\t" << p[0];
                p += matrix.stride(1);
            }

            stream << std::endl;
        }
        stream << "]" << std::endl;
    }

    return stream;
}

template <typename SrcType, typename DstType, size_t Channels, typename Func, typename ...Args>
void processElementwise(const Matrix<SrcType, Channels> &src, Matrix<DstType, Channels> &dst, Func f, Args&&... args)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t srcStrideX = src.stride(1);

    dst.resize(rows, cols);
    size_t dstStrideX = dst.stride(1);

    if (src.isContinuous())
    {
        cols = cols * rows;
        rows = 1;
    }

    for (size_t y = 0; y < rows; ++y)
    {
        const SrcType *ptrSrc = src.ptr(y);
        DstType *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = f(ptrSrc[c], std::forward<Args>(args)...);
            }
            ptrSrc += srcStrideX;
            ptrDst += dstStrideX;
        }
    }
}

template <typename SrcType0, typename SrcType1, typename DstType, size_t Channels, typename Func, typename ...Args>
void processElementwise(const Matrix<SrcType0, Channels> &src0, const Matrix<SrcType1, Channels> &src1, Matrix<DstType, Channels> &dst, Func f, Args&&... args)
{
    size_t rows = src0.rows();
    size_t cols = src0.cols();
    size_t src0StrideX = src0.stride(1);
    size_t src1StrideX = src1.stride(1);

    if (rows != src1.rows() || cols != src1.cols())
    {
        auto errMsg = std::ostringstream{};
        errMsg << "Matrix size does not match. ";
        errMsg << rows << "x" << cols << " != " << src1.rows() << "x" << src1.cols();
        throw std::invalid_argument(errMsg.str());
    }

    dst.resize(rows, cols);
    size_t dstStrideX = dst.stride(1);

    if (src0.isContinuous() && src1.isContinuous())
    {
        cols = cols * rows;
        rows = 1;
    }

    for (size_t y = 0; y < rows; ++y)
    {
        const auto *ptrSrc0 = src0.ptr(y);
        const auto *ptrSrc1 = src1.ptr(y);
        DstType *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = f(ptrSrc0[c], ptrSrc1[c], std::forward<Args>(args)...);
            }
            ptrSrc0 += src0StrideX;
            ptrSrc1 += src1StrideX;
            ptrDst += dstStrideX;
        }
    }
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator +(const Matrix<T, Channels> &src, const U &val)
{
    Matrix<T, Channels> dst;
    processElementwise(src, dst, [&](const auto &s) { return s + val; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator -(const Matrix<T, Channels> &src, const U &val)
{
    Matrix<T, Channels> dst;
    processElementwise(src, dst, [&](const auto &s) { return s - val; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator *(const Matrix<T, Channels> &src, const U &val)
{
    Matrix<T, Channels> dst;
    processElementwise(src, dst, [&](const auto &s) { return s * val; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator /(const Matrix<T, Channels> &src, const U &val)
{
    Matrix<T, Channels> dst;
    processElementwise(src, dst, [&](const auto &s) { return s / val; });
    return dst;
}


template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator +(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1)
{
    Matrix<T, Channels> dst;
    processElementwise(src0, src1, dst, [&](const auto &s0, const auto &s1) { return s0 + s1; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator -(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1)
{
    Matrix<T, Channels> dst;
    processElementwise(src0, src1, dst, [&](const auto &s0, const auto &s1) { return s0 - s1; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator *(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1)
{
    Matrix<T, Channels> dst;
    processElementwise(src0, src1, dst, [&](const auto &s0, const auto &s1) { return s0 * s1; });
    return dst;
}

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator /(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1)
{
    Matrix<T, Channels> dst;
    processElementwise(src0, src1, dst, [&](const auto &s0, const auto &s1) { return s0 / s1; });
    return dst;
}

template<typename T, size_t Channels, typename Compare>
inline Matrix<uint8_t, Channels> compare(const Matrix<T, Channels> &src, T val, Compare comp)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t srcStrideX = src.stride(1);

    Matrix<uint8_t, Channels> dst(rows, cols);
    size_t dstStrideX = dst.stride(1);

    if (src.isContinuous())
    {
        cols = cols * rows;
        rows = 1;
    }

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        uint8_t *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = comp(ptrSrc[c], val) ? 255 : 0;
            }
            ptrSrc += srcStrideX;
            ptrDst += dstStrideX;
        }
    }

    return dst;
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator <(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::less<T>());
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator <=(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::less_equal<T>());
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator >(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::greater<T>());
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator >=(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::greater_equal<T>());
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator ==(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::equal_to<T>());
}

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator !=(const Matrix<T, Channels> &src, T val)
{
    return compare(src, val, std::not_equal_to<T>());
}

template <typename T, size_t Channels>
static Matrix<T, Channels> merge(const Vec<Matrix<T, 1>, Channels> &channels)
{
    size_t rows = channels[0].rows();
    size_t cols = channels[0].cols();

    Matrix<T, Channels> dst(rows, cols);
    size_t dstStrideX = dst.stride(1);

    for (size_t c = 0; c < Channels; ++c)
    {
        size_t srcStrideX = channels[c].stride(1);

        for (size_t y = 0; y < rows; ++y)
        {
            const T *ptrSrc = channels[c].ptr(y);
            T *ptrDst = dst.ptr(y) + c;

            for (size_t x = 0; x < cols; ++x)
            {
                ptrDst[0] = ptrSrc[0];
                ptrSrc += srcStrideX;
                ptrDst += dstStrideX;
            }
        }
    }

    return dst;
}

template <typename T, size_t Channels>
static Matrix<T, Channels> swapChannels(const Matrix<T, Channels> &src)
{
    size_t rows = src.rows();
    size_t cols = src.cols();
    size_t srcStrideX = src.stride(1);

    Matrix<T, Channels> dst(rows, cols);
    size_t dstStrideX = dst.stride(1);

    if (src.isContinuous())
    {
        cols = cols * rows;
        rows = 1;
    }

    for (size_t y = 0; y < rows; ++y)
    {
        const T *ptrSrc = src.ptr(y);
        T *ptrDst = dst.ptr(y);

        for (size_t x = 0; x < cols; ++x)
        {
            for (size_t c = 0; c < Channels; ++c)
            {
                ptrDst[c] = ptrSrc[Channels-c-1];
            }
            ptrSrc += srcStrideX;
            ptrDst += dstStrideX;
        }
    }

    return dst;
}

}

#endif
