/** \file
 * \brief Matrix classes
 */

#ifndef OSP_MAT_H
#define OSP_MAT_H

#include <array>
#include <atomic>
#include <cassert>
#include <cstring>
#include <iostream>
#include <memory>
#include <mutex>
#include <sstream>
#include <osp/types/buffer.h>

namespace osp {

template <typename T, size_t Rows, size_t Cols>
class Mat
{
public:
    Mat();
    Mat(std::initializer_list<T> initializer);
    Mat(const Mat &other);
    Mat(const T *data);

    friend void swap(Mat &a, Mat &b)
    {
        using std::swap;
        swap(a.m_data, b.m_data);
    }

    Mat &operator=(Mat other);

    static Mat eye();

    size_t rows() const;
    size_t cols() const;

    void fill(const T &val);
    typename std::array<T, Rows*Cols>::reference &at(size_t y, size_t x);
    typename std::array<T, Rows*Cols>::const_reference &at(size_t y, size_t x) const;

    typename std::array<T, Rows*Cols>::reference operator[](size_t idx);
    typename std::array<T, Rows*Cols>::const_reference operator[](size_t idx) const;

    typename std::array<T, Rows*Cols>::pointer data();
    typename std::array<T, Rows*Cols>::const_pointer data() const;

private:
    std::array<T, Rows*Cols> m_data;
};

template <typename SrcType, typename DstType, size_t Rows, size_t Cols, typename Func, typename ...Args>
void processElementwise(const Mat<SrcType, Rows, Cols> &src, Mat<DstType, Rows, Cols> &dst, Func f, Args&&... args);

template <typename SrcType0, typename SrcType1, typename DstType, size_t Rows, size_t Cols, typename Func, typename ...Args>
void processElementwise(const Mat<SrcType0, Rows, Cols> &src0, const Mat<SrcType1, Rows, Cols> &src1, Mat<DstType, Rows, Cols> &dst, Func f, Args&&... args);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator*(const Mat<T, Rows, Cols> &src, const U &val);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const Mat<T, Rows, Cols> &src, const U &val);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator-(const Mat<T, Rows, Cols> &src, const U &val);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator/(const Mat<T, Rows, Cols> &src, const U &val);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator*(const U &val, const Mat<T, Rows, Cols> &src);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const U &val, const Mat<T, Rows, Cols> &src);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator+(const Mat<T, Rows, Cols> &src0, const Mat<U, Rows, Cols> &src1);

template <typename T, typename U, size_t Rows, size_t Cols>
Mat<T, Rows, Cols> operator-(const Mat<T, Rows, Cols> &src0, const Mat<U, Rows, Cols> &src1);



template<typename T, size_t Size>
using Vec = Mat<T, Size, 1>;

template<typename T>
using Vec2 = Vec<T, 2>;

template<typename T>
using Vec3 = Vec<T, 3>;

template <size_t N, class T>
Vec<T ,N> makeVec(const T &v);


/*!
 * \brief The Matrix class defines a MxN matrix with a fixed number of channels.
 */
template<typename T, size_t Channels = 1u>
class Matrix
{
public:
    Matrix();
    Matrix(size_t rows, size_t cols, T val = 0);
    Matrix(size_t rows, size_t cols, Vec<T, Channels> val);
    Matrix(size_t rows, size_t cols, T *data, Vec<size_t, 2> stride = {0, 0}, size_t offset = 0);
    Matrix(size_t rows, size_t cols, Buffer<T> buf, Vec<size_t, 2> stride = {0, 0}, size_t offset = 0);
    Matrix(const Matrix<T, Channels> & other);
    Matrix(Matrix<T, Channels> && other);

    Matrix<T, Channels> & operator =(Matrix<T, Channels> other);
    void swap(Matrix<T, Channels> & other) noexcept;

    void resize(size_t rows, size_t cols);

    template <typename DstType>
    Matrix<DstType, Channels> convert() const;

    Matrix<T, Channels> transpose() const;

    T & at(size_t y, size_t x, size_t channel = 0);
    const T & at(size_t y, size_t x, size_t channel = 0) const;

    size_t rows() const;
    size_t cols() const;
    size_t channels() const;
    Vec<size_t, 2> stride() const;
    size_t stride(size_t idx) const;

    bool empty() const;
    size_t total() const;
    size_t elemSize() const;
    size_t bytes() const;

    Buffer<T> internalBuffer() const;

    T* ptr(size_t y = 0) const;
    T* ptr(size_t y, size_t x) const;
    T* ptr(size_t y, size_t x, size_t channel) const;

    bool isContinuous() const;
    Matrix &makeContinuous();

    Matrix clone() const;

    void fill(T val);
    void fill(Vec<T, Channels> val);
    template <typename U>
    void setTo(T val, Matrix<U, 1> mask);
    template <typename U>
    void setTo(Vec<T, Channels> val, Matrix<U, 1> mask);
    Matrix<T, 1> channel(size_t c) const;

private:
    size_t m_rows;
    size_t m_cols;
    Vec<size_t, 2> m_stride;
    size_t m_offset;
    bool m_continuous;
    Buffer<T> m_buf;
};

enum class Depth
{
    Int8,
    Int16,
    Int32,
    Int64,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    Float,
    Double
};

template<typename T>
struct DataType {};
template<> struct DataType<int8_t> { static const Depth depth = Depth::Int8; };
template<> struct DataType<int16_t> { static const Depth depth = Depth::Int16; };
template<> struct DataType<int32_t> { static const Depth depth = Depth::Int32; };
template<> struct DataType<int64_t> { static const Depth depth = Depth::Int64; };
template<> struct DataType<uint8_t> { static const Depth depth = Depth::UInt8; };
template<> struct DataType<uint16_t> { static const Depth depth = Depth::UInt16; };
template<> struct DataType<uint32_t> { static const Depth depth = Depth::UInt32; };
template<> struct DataType<uint64_t> { static const Depth depth = Depth::UInt64; };
template<> struct DataType<float> { static const Depth depth = Depth::Float; };
template<> struct DataType<double> { static const Depth depth = Depth::Double; };

template<Depth depth>
struct DepthType {};
template<> struct DepthType<Depth::Int8> { using valueType = int8_t; };
template<> struct DepthType<Depth::Int16> { using valueType = int16_t; };
template<> struct DepthType<Depth::Int32> { using valueType = int32_t; };
template<> struct DepthType<Depth::Int64> { using valueType = int64_t; };
template<> struct DepthType<Depth::UInt8> { using valueType = int8_t; };
template<> struct DepthType<Depth::UInt16> { using valueType = int16_t; };
template<> struct DepthType<Depth::UInt32> { using valueType = int32_t; };
template<> struct DepthType<Depth::UInt64> { using valueType = int64_t; };
template<> struct DepthType<Depth::Float> { using valueType = float; };
template<> struct DepthType<Depth::Double> { using valueType = double; };

template<typename T, size_t Rows, size_t Cols>
std::ostream &operator <<(std::ostream &stream, const Mat<T, Rows, Cols> &matrix);
template<typename T, size_t Channels>
std::ostream &operator <<(std::ostream &stream, const Matrix<T, Channels> &matrix);

template <typename SrcType, typename DstType, size_t Channels, typename Func, typename ...Args>
void processElementwise(const Matrix<SrcType, Channels> &src, Matrix<DstType, Channels> &dst, Func f, Args&&... args);

template <typename SrcType0, typename SrcType1, typename DstType, size_t Channels, typename Func, typename ...Args>
void processElementwise(const Matrix<SrcType0, Channels> &src0, const Matrix<SrcType1, Channels> &src1, Matrix<DstType, Channels> &dst, Func f, Args&&... args);

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator +(const Matrix<T, Channels> &src, const U &val);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator -(const Matrix<T, Channels> &src, const U &val);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator *(const Matrix<T, Channels> &src, const U &val);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator /(const Matrix<T, Channels> &src, const U &val);

template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator +(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator -(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator *(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1);
template<typename T, typename U, size_t Channels>
inline Matrix<T, Channels> operator /(const Matrix<T, Channels> &src0, const Matrix<U, Channels> &src1);

template<typename T, size_t Channels, typename Compare>
inline Matrix<uint8_t, Channels> compare(const Matrix<T, Channels> &src, T val, Compare comp);

template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator <(const Matrix<T, Channels> &src, T val);
template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator <=(const Matrix<T, Channels> &src, T val);
template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator >(const Matrix<T, Channels> &src, T val);
template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator >=(const Matrix<T, Channels> &src, T val);
template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator ==(const Matrix<T, Channels> &src, T val);
template<typename T, size_t Channels>
inline Matrix<uint8_t, Channels> operator !=(const Matrix<T, Channels> &src, T val);

template <typename T, size_t Channels>
static Matrix<T, Channels> merge(const Vec<Matrix<T, 1>, Channels> &channels);
template <typename T, size_t Channels>
static Matrix<T, Channels> swapChannels(const Matrix<T, Channels> &src);

}

#include <osp/types/details/mat.h>

#endif // OSP_MAT_H
