#ifndef TS_MATHELPER_H
#define TS_MATHELPER_H

#include <opencv2/core/core.hpp>
#include <osp/types/coordinates.h>
#include <osp/types/mat.h>
#include <osp/types/ndarray.h>

namespace osp {

template<typename T, size_t Channels>
inline cv::Mat matrixToCvMat(Matrix<T, Channels> &matrix, bool clone = false)
{
    if (matrix.stride(1) != Channels)
        throw std::runtime_error("Stride x must be equal to the number of channels. Clone matrix first.");

    cv::Mat cvMat(matrix.rows(),
                  matrix.cols(),
                  CV_MAKETYPE(cv::DataType<T>::type, Channels),
                  matrix.ptr(),
                  matrix.stride(0) * matrix.elemSize());

    if (clone)
        return cvMat.clone();
    else
        return cvMat;
}

template<typename T, size_t Channels>
inline const cv::Mat matrixToCvMat(const Matrix<T, Channels> &matrix, bool clone = false)
{
    if (matrix.stride(1) != Channels)
        throw std::runtime_error("Stride x must be equal to the number of channels. Clone matrix first.");

    cv::Mat cvMat(matrix.rows(),
                  matrix.cols(),
                  CV_MAKETYPE(cv::DataType<T>::type, Channels),
                  matrix.ptr(),
                  matrix.stride(0) * matrix.elemSize());

    if (clone)
        return cvMat.clone();
    else
        return cvMat;
}

template<typename T, size_t Channels>
inline cv::Mat matrixToCvMat(Matrix<T, Channels> &&matrix)
{
    return matrixToCvMat(matrix, true);
}

template<typename T, size_t Channels>
inline Matrix<T, Channels> cvMatToMatrix(cv::Mat &cvMat, bool clone = true)
{
    if (cvMat.type() != CV_MAKETYPE(cv::DataType<T>::type, Channels))
        throw std::invalid_argument("cv::Mat has wrong data type.");

    Buffer<T> buf(cvMat.rows*cvMat.cols*Channels, cvMat.ptr<T>());
    Matrix<T, Channels> matrix(cvMat.rows, cvMat.cols, buf);

    if (clone)
        return matrix.clone();
    else
        return matrix;
}

template<typename T, size_t Channels>
inline Matrix<T, Channels> cvMatToMatrix(const cv::Mat &cvMat)
{
    if (cvMat.type() != CV_MAKETYPE(cv::DataType<T>::type, Channels))
        throw std::invalid_argument("cv::Mat has wrong data type.");

    Buffer<T> buf(cvMat.rows*cvMat.cols*Channels, const_cast<T *>(cvMat.ptr<const T>()));
    Matrix<T, Channels> matrix(cvMat.rows, cvMat.cols, buf);

    return matrix.clone();
}

template<typename T, size_t Channels>
inline Matrix<T, Channels> cvMatToMatrix(cv::Mat &&cvMat)
{
    return cvMatToMatrix<T, Channels>(cvMat, true);
}

template <typename T>
NDArray<T> cvMatToNDArray(cv::Mat &cvMat, bool clone = false)
{
    if (cvMat.depth() != cv::DataType<T>::depth)
        throw std::invalid_argument("cv::Mat has wrong data type.");

    NDArray<T> array({static_cast<size_t>(cvMat.rows),
                      static_cast<size_t>(cvMat.cols),
                      static_cast<size_t>(cvMat.channels())},
                     cvMat.ptr<T>(),
                     {cvMat.step1(0), cvMat.step1(1), 1u});

    if (clone)
        return array.clone();
    else
        return array;
}

template <typename T>
cv::Mat ndArrayToCvMat(NDArray<T> &array, bool clone = false)
{
    auto dims = array.dims();

    cv::Mat mat;
    if (dims.size() <= 3)
        mat = cv::Mat(dims[0], dims[1], CV_MAKETYPE(cv::DataType<T>::type, dims[2]), array.data());
    else
        mat = cv::Mat(std::vector<int>(dims.begin(), dims.end()), cv::DataType<T>::type, array.data());

    if (clone)
        return mat.clone();
    else
        return mat;
}

template <typename T, size_t Channels>
bool convertible(const cv::Mat &src, const Matrix<T, Channels> &/*dst*/)
{
    return src.type() == CV_MAKETYPE(cv::DataType<T>::type, Channels);
}

template <typename T>
bool convertible(const cv::Mat &src, const Matrix<Point2d<T>> &/*dst*/)
{
    return src.type() == CV_MAKETYPE(cv::DataType<T>::type, 2);
}

template <typename T>
bool convertible(const cv::Mat &src, const Matrix<Polar<T>> &/*dst*/)
{
    return src.type() == CV_MAKETYPE(cv::DataType<T>::type, 2);
}

template <typename T>
bool convertible(const cv::Mat &src, const Matrix<Point3d<T>> &/*dst*/)
{
    return src.type() == CV_MAKETYPE(cv::DataType<T>::type, 3);
}

template <typename T>
bool convertible(const cv::Mat &src, const Matrix<Spherical<T>> &/*dst*/)
{
    return src.type() == CV_MAKETYPE(cv::DataType<T>::type, 3);
}

template <typename T, size_t Channels>
void convert(cv::Mat &&src, Matrix<T, Channels> &dst)
{
    if (!convertible(src, dst))
        throw std::invalid_argument("cv::Mat has wrong data type.");

    Buffer<T> buf(src.rows*src.cols*Channels, const_cast<T*>(src.ptr<T>()));
    Matrix<T, Channels> matrix(src.rows, src.cols, buf);

    dst = matrix.clone();
}

template <typename T, size_t Channels>
void convert(const cv::Mat &src, Matrix<T, Channels> &dst, bool copy = true)
{
    if (!convertible(src, dst))
        throw std::invalid_argument("cv::Mat has wrong data type.");

    Buffer<T> buf(src.rows*src.cols*Channels, const_cast<T*>(src.ptr<T>()));
    Matrix<T, Channels> matrix(src.rows, src.cols, buf);

    if (copy)
        dst = matrix.clone();
    else
        dst = matrix;
}

template<typename T, size_t Channels>
int cvTypeFromMatrix(const Matrix<T, Channels> &/*mat*/)
{
    return CV_MAKETYPE(cv::DataType<T>::type, Channels);
}

template<typename T>
int cvTypeFromMatrix(const Matrix<Point2d<T>> &/*mat*/)
{
    return CV_MAKETYPE(cv::DataType<T>::type, 2);
}

template<typename T>
int cvTypeFromMatrix(const Matrix<Polar<T>> &/*mat*/)
{
    return CV_MAKETYPE(cv::DataType<T>::type, 2);
}

template<typename T>
int cvTypeFromMatrix(const Matrix<Point3d<T>> &/*mat*/)
{
    return CV_MAKETYPE(cv::DataType<T>::type, 3);
}

template<typename T>
int cvTypeFromMatrix(const Matrix<Spherical<T>> &/*mat*/)
{
    return CV_MAKETYPE(cv::DataType<T>::type, 3);
}

template<typename T, size_t Channels>
void convert(Matrix<T, Channels> &&src, cv::Mat &dst)
{
    if (src.stride(1) != Channels)
        throw std::runtime_error("Stride x must be equal to the number of channels. Clone matrix first.");

    cv::Mat cvMat(src.rows(),
                  src.cols(),
                  cvTypeFromMatrix(src),
                  src.ptr(),
                  src.stride(0) * src.elemSize());

    dst = cvMat.clone();
}

template<typename T, size_t Channels>
void convert(const Matrix<T, Channels> &src, cv::Mat &dst, bool clone = false)
{
    if (src.stride(1) != Channels)
        throw std::runtime_error("Stride x must be equal to the number of channels. Clone matrix first.");

    cv::Mat cvMat(src.rows(),
                  src.cols(),
                  cvTypeFromMatrix(src),
                  src.ptr(),
                  src.stride(0) * src.elemSize());

    if (clone)
        dst = cvMat.clone();
    else
        dst = cvMat;
}

template <typename T, size_t Channels = 1>
Matrix<T, Channels> convertToMatrix(cv::Mat &&src)
{
    Matrix<T, Channels> dst;
    convert(src, dst);
    return dst;
}

template <typename T, size_t Channels>
cv::Mat convertToCvMat(Matrix<T, Channels> &&src)
{
    cv::Mat dst;
    convert(src, dst);
    return dst;
}

template <typename T, size_t Channels = 1>
Matrix<T, Channels> convertToMatrix(const cv::Mat &src, bool copy = true)
{
    Matrix<T, Channels> dst;
    convert(src, dst, copy);
    return dst;
}

template <typename T, size_t Channels>
cv::Mat convertToCvMat(const Matrix<T, Channels> &src, bool copy = true)
{
    cv::Mat dst;
    convert(src, dst, copy);
    return dst;
}

}

#endif // MAT_H
