#ifndef OSP_TYPES_NDARRAY_H
#define OSP_TYPES_NDARRAY_H

#include <memory>
#include <numeric>
#include <vector>
#include "memory.h"

namespace osp
{

template <typename T, typename MemoryTag = HostMemory>
class NDArray
{
public:
    using ValueType = T;
    using PtrType = ValueType *;
    using MemoryType = Memory<T, MemoryTag>;

    NDArray() = default;

    template <typename OtherMemoryTag>
    NDArray(const NDArray<T, OtherMemoryTag> &other);

    NDArray(std::vector<std::size_t> size);
    NDArray(std::vector<std::size_t> dims,
            PtrType data,
            std::vector<std::size_t> stride = std::vector<std::size_t>(),
            bool takeOwnership = false);

    void resize(std::vector<std::size_t> dims);
    void reshape(std::vector<std::size_t> dims);
    void reset();
    std::vector<std::size_t> dims() const;
    std::size_t ndims() const;
    std::vector<std::size_t> stride() const;
    bool continuous() const;
    bool empty() const;

    PtrType data();
    PtrType data() const;

    PtrType ptr(size_t i = 0);
    PtrType ptr(size_t i, size_t j);
    PtrType ptr(size_t i, size_t j, size_t k);
    PtrType ptr(std::vector<std::size_t> dims);

    NDArray clone() const;

    std::shared_ptr<MemoryType> mem();
    const std::shared_ptr<MemoryType> mem() const;

private:
    void allocate();
    void setDefaultStride();

    std::vector<std::size_t> m_dims;
    std::vector<std::size_t> m_stride;
    std::shared_ptr<MemoryType> m_memory;
    bool m_continuous;
};

template <typename T, typename MemoryTag>
template <typename OtherMemoryTag>
NDArray<T, MemoryTag>::NDArray(const NDArray<T, OtherMemoryTag> &other)
{
    m_dims = other.dims();
    allocate();
    copyMemory(*other.mem(), *m_memory);
}

template <typename T, typename MemoryTag>
NDArray<T, MemoryTag>::NDArray(std::vector<std::size_t> dims)
{
    resize(dims);
}

template <typename T, typename MemoryTag>
NDArray<T, MemoryTag>::NDArray(std::vector<std::size_t> dims, PtrType data, std::vector<std::size_t> stride, bool takeOwnership)
{
    m_dims = dims;

    if (stride.empty())
    {
        setDefaultStride();
    }
    else
    {
        m_stride = stride;
        if (m_dims[0]*m_stride[0] != std::accumulate(m_dims.begin(), m_dims.end(), 1u, std::multiplies<std::size_t>()))
            m_continuous = false;
    }

    m_memory = std::make_shared<MemoryType>(data,
                                            m_stride[0]*m_dims[0],
                                            takeOwnership);
}

template <typename T, typename MemoryTag>
void NDArray<T, MemoryTag>::resize(std::vector<std::size_t> dims)
{
    m_dims = dims;
    allocate();
}

template <typename T, typename MemoryTag>
void NDArray<T, MemoryTag>::reshape(std::vector<std::size_t> dims)
{
    if (std::accumulate(dims.begin(), dims.end(), 1, std::multiplies<std::size_t>()) !=
            std::accumulate(m_dims.begin(), m_dims.end(), 1, std::multiplies<std::size_t>()))
    m_dims = dims;
}

template <typename T, typename MemoryTag>
void NDArray<T, MemoryTag>::reset()
{
    m_dims.resize(0);
    m_memory.reset();
}

template <typename T, typename MemoryTag>
std::vector<std::size_t> NDArray<T, MemoryTag>::dims() const
{
    return m_dims;
}

template <typename T, typename MemoryTag>
std::size_t NDArray<T, MemoryTag>::ndims() const
{
    return m_dims.size();
}

template <typename T, typename MemoryTag>
std::vector<std::size_t> NDArray<T, MemoryTag>::stride() const
{
    return m_stride;
}

template <typename T, typename MemoryTag>
bool NDArray<T, MemoryTag>::continuous() const
{
    return m_continuous;
}

template <typename T, typename MemoryTag>
bool NDArray<T, MemoryTag>::empty() const
{
    return m_memory == nullptr || m_memory->size() == 0;
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::data()
{
    if (!m_memory)
        return nullptr;
    return m_memory->data();
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::data() const
{
    if (!m_memory)
        return nullptr;
    return m_memory->data();
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::ptr(size_t i)
{
    return ptr(std::vector<std::size_t>{i});
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::ptr(size_t i, size_t j)
{
    return ptr({i, j});
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::ptr(size_t i, size_t j, size_t k)
{
    return ptr({i, j, k});
}

template <typename T, typename MemoryTag>
typename NDArray<T, MemoryTag>::PtrType NDArray<T, MemoryTag>::ptr(std::vector<std::size_t> dims)
{
    if (dims.size() > m_stride.size())
        throw std::out_of_range("Array has only " + std::to_string(m_stride.size()) + " dimensions.");

    if (!m_memory)
        return nullptr;
    return m_memory->data() + std::inner_product(dims.begin(), dims.end(), m_stride.begin(), 0);
}

template <typename T, typename MemoryTag>
NDArray<T, MemoryTag> NDArray<T, MemoryTag>::clone() const
{
    return NDArray<T, MemoryTag>(*this);
}

template <typename T, typename MemoryTag>
std::shared_ptr<typename NDArray<T, MemoryTag>::MemoryType> NDArray<T, MemoryTag>::mem()
{
    return m_memory;
}

template <typename T, typename MemoryTag>
const std::shared_ptr<typename NDArray<T, MemoryTag>::MemoryType> NDArray<T, MemoryTag>::mem() const
{
    return m_memory;
}

template <typename T, typename MemoryTag>
void NDArray<T, MemoryTag>::allocate()
{
    m_memory = std::make_shared<MemoryType>(
        std::accumulate(m_dims.begin(), m_dims.end(), 1, std::multiplies<std::size_t>()));
    setDefaultStride();
}

template <typename T, typename MemoryTag>
void NDArray<T, MemoryTag>::setDefaultStride()
{
    std::vector<std::size_t> stride(m_dims.size());
    stride[static_cast<int>(stride.size())-1] = 1;
    for (int i = static_cast<int>(stride.size())-2; i >= 0; --i)
        stride[i] = stride[i+1] * m_dims[i+1];
    m_stride = stride;
    m_continuous = true;
}

} // namespace osp

#endif
