#include <atomic>
#include <chrono>
#include <csignal>
#include <iomanip>
#include <thread>
#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#include <osp/sdk/device.h>
#include <osp/types/mathelper.h>

#ifdef OPENCV_STATIC_BUILD
#include <QtCore/qglobal.h>
#endif

namespace config {
    std::string device;
    bool tls;
    osp::Device::CaptureMode captureMode;
    std::string outputDir;
    int interval;
}

namespace po = boost::program_options;

std::atomic_bool running;

void signalHandler(int signal)
{
    running = false;
}

bool parseCmdLineArgs(int argc, char **argv)
{
    // Declare the supported options.
    po::options_description desc("Allowed options");
    desc.add_options()
        ("help", "produce help message")
        ("device", po::value<std::string>()->default_value("auto"), "IP or Hostname, use \"auto\" for device discovery")
        ("tls", "enable tls encryption")
        ("mode", po::value<std::string>()->default_value("rgbd"), "set mode (rgbd/raw)")
        ("output_dir", po::value<std::string>()->default_value("."), "output directory")
        ("interval", po::value<int>()->default_value(2000), "recording interval in seconds")
    ;

    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, desc), vm);
    po::notify(vm);

    if (vm.count("help"))
    {
        std::cout << desc << std::endl;
        return false;
    }

    config::tls = vm.count("tls");

    if (vm.count("device"))
    {
        config::device = vm["device"].as<std::string>();
    }

    if (vm.count("mode"))
    {
        if (vm["mode"].as<std::string>() == "rgbd")
        {
            config::captureMode = osp::Device::CaptureMode::RGBD;
        }
        else if (vm["mode"].as<std::string>() == "raw")
        {
            config::captureMode = osp::Device::CaptureMode::RAW;
        }
        else
        {
            std::cout << "Wrong mode. Please set mode to \"raw\" or \"rgbd\".";
            return false;
        }
    }
    
    config::outputDir = vm["output_dir"].as<std::string>();
    config::interval = vm["interval"].as<int>();
    
    return true;
}

void saveImage(const std::string &filepath, const osp::Matrix<uint8_t, 3> &image)
{
    auto cvMatRgb = osp::matrixToCvMat(image);
    cv::Mat cvMatBgr;
    cv::cvtColor(cvMatRgb, cvMatBgr, cv::COLOR_RGB2BGR);
    cv::imwrite(filepath, cvMatBgr);
}

void saveDistanceMap(const std::string &filepath, const osp::Matrix<uint16_t, 1> &distanceMap)
{
    cv::Mat distanceMapCv = osp::matrixToCvMat(distanceMap);
    cv::Mat distanceMapColored;
    distanceMapCv.convertTo(distanceMapColored, CV_8U, 255./5000.);

    cv::applyColorMap(255-distanceMapColored, distanceMapColored, cv::COLORMAP_JET);
    cv::imwrite(filepath, distanceMapColored);
}

template <typename T>
std::string numToStr(T num, int size = 5)
{
    std::stringstream ss;
    ss << std::setw(size) << std::setfill('0') << num;
    return ss.str();
}

void process(osp::Device *device)
{
    size_t frameCount = 0;

    running = true;
    while(running)
    {
        auto frame = device->getFrame();

        if (!frame)
            break;

        std::string frameCountStr = numToStr(frameCount);

        if (!frame->image.empty())
        {
            boost::filesystem::create_directories(config::outputDir + "/image");
            saveImage(config::outputDir + "/image/" + frameCountStr + ".png", frame->image);
        }

        if (!frame->distanceMap.empty())
        {
            boost::filesystem::create_directories(config::outputDir + "/distance");
            saveDistanceMap(config::outputDir + "/distance/" + frameCountStr + ".png", frame->distanceMap);
        }

        if (!frame->sourceImage0.empty())
        {
            boost::filesystem::create_directories(config::outputDir + "/image_0");
            saveImage(config::outputDir + "/image_0/" + frameCountStr + ".png", frame->sourceImage0);
        }

        if (!frame->sourceImage1.empty())
        {
            boost::filesystem::create_directories(config::outputDir + "/image_1");
            saveImage(config::outputDir + "/image_1/" + frameCountStr + ".png", frame->sourceImage1);
        }

        if (!frame->sourceImage2.empty())
        {
            boost::filesystem::create_directories(config::outputDir + "/image_2");
            saveImage(config::outputDir + "/image_2/" + frameCountStr + ".png", frame->sourceImage2);
        }
        
        std::cout << "Recorded image " << frameCountStr << std::endl;

        ++frameCount;
        std::this_thread::sleep_for(std::chrono::milliseconds(config::interval));
    }
}

int main(int argc, char **argv)
{
    if (!parseCmdLineArgs(argc, argv))
        return EXIT_FAILURE;


    std::unique_ptr<osp::Device> device;
    if (config::device == "auto")
    {
        // scan available devices
        std::vector<osp::Device::Info> devices = osp::DeviceManager::get().getDevices();
        if (devices.empty())
        {
            std::cout << "No devices found." << std::endl;
            return EXIT_FAILURE;
        }

        // connect to the first device
        device = osp::DeviceManager::get().request(devices[0], config::tls);
    }
    else
    {
        // connect to device
        device = osp::DeviceManager::get().request(config::device, config::tls);
    }

    if (!device)
    {
        std::cerr << "Device " << config::device << " not available." << std::endl;
        return EXIT_FAILURE;
    }

    // login if password is required
    auto info = device->info();
    if (info.passwordRequired)
    {
        osp::PasswordCredentials creds;
        std::cout << "Password: ";
        std::cin >> creds.password;
        auto loginStatus = device->login(creds);
        if (!loginStatus.ok())
        {
            std::cout << "Login failed!" << std::endl;
            return EXIT_FAILURE;
        }
    }

    // set capture mode
    device->setCaptureMode(config::captureMode);

    // start streaming from device
    osp::Status status = device->start();

    if (!status.ok())
        return EXIT_FAILURE;
        
    // register signal handler
    std::signal(SIGTERM, signalHandler);
    std::signal(SIGINT, signalHandler);

    // start main loop
    process(device.get());

    // stop streaming from device
    device->stop();

    // close device connection
    device->close();

    return EXIT_SUCCESS;
}
