#include <osp/cv/vptz.h>
#include <osp/sdk/device.h>
#include <osp/types/mathelper.h>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>


using namespace osp;

static VPtz vptz;

void showDistance(const std::string &winName, Matrix<uint16_t, 1> distance)
{
    cv::Mat distanceCv(distance.rows(), distance.cols(), CV_16U, distance.ptr());
    cv::Mat distanceCvColored;
    distanceCv.convertTo(distanceCvColored, CV_8U, 255./5000.);
    cv::imshow(winName, distanceCvColored);
}

void process(osp::Device *device)
{
    char key;
    do
    {
        auto frame = device->getFrame();
        
        if (!frame)
            break;
        
        if (!frame->image.empty())
        {
            cv::namedWindow("Image", cv::WINDOW_NORMAL);
            cv::imshow("Image", cv::Mat(frame->image.rows(), frame->image.cols(), CV_8UC3, frame->image.ptr()));

            auto imgVirt = vptz.transform(frame->image);
            cv::namedWindow("Image VPTZ", cv::WINDOW_NORMAL);
            cv::imshow("Image VPTZ", cv::Mat(imgVirt.rows(), imgVirt.cols(), CV_8UC3, imgVirt.ptr()));
        }

        if (!frame->distanceMap.empty())
        {
            cv::namedWindow("Distance", cv::WINDOW_NORMAL);
            showDistance("Distance", frame->distanceMap);

            auto distVirt = vptz.transform(frame->distanceMap);
            cv::namedWindow("Distance VPTZ", cv::WINDOW_NORMAL);
            showDistance("Distance VPTZ", distVirt);
        }
        
        key = cv::waitKey(10);

        switch (key)
        {
        case 'w':
            vptz.setTilt(vptz.tilt() + 0.1);
            break;
        case 's':
            vptz.setTilt(vptz.tilt() - 0.1);
            break;
        case 'd':
            vptz.setPan(vptz.pan() + 0.1);
            break;
        case 'a':
            vptz.setPan(vptz.pan() - 0.1);
            break;
        case 'e':
            vptz.setRoll(vptz.roll() + 0.1);
            break;
        case 'q':
            vptz.setRoll(vptz.roll() - 0.1);
            break;
        case '+':
            vptz.setFocalLength(vptz.focalLength() + 20);
            break;
        case '-':
            vptz.setFocalLength(vptz.focalLength() - 20);
            break;
        }   
    }
    while(key != 'x');
}

int main()
{
    // scan available devices
    std::vector<osp::Device::Info> devices = osp::DeviceManager::get().getDevices();
    if (devices.empty())
    {
        std::cout << "No devices found." << std::endl;
        return EXIT_FAILURE;
    }

    // connect to the first device
    std::unique_ptr<osp::Device> device = osp::DeviceManager::get().request(devices[0]);

    // login if password is required
    auto info = device->info();
    if (info.passwordRequired)
    {
        osp::PasswordCredentials creds;
        std::cout << "Password: ";
        std::cin >> creds.password;
        auto loginStatus = device->login(creds);
        if (!loginStatus.ok())
        {
            std::cout << "Login failed!" << std::endl;
            return EXIT_FAILURE;
        }
    }

    // configure virtual pan-tilt-zoom camera
    vptz.setSourceCalib(device->getCameraModel(CameraModelType::Target));
    vptz.setFocalLength(200);
    vptz.setImageResolution(640, 480);
    vptz.setOrientation(0, 0, 0);

    // set capture mode
    device->setCaptureMode(Device::CaptureMode::RGBD);

    // start streaming from device
    osp::Status status = device->start();

    process(device.get());

    device->stop();
    device->close();

    return EXIT_SUCCESS;
}
